from __future__ import annotations

import contextlib
from collections.abc import Mapping
from collections import defaultdict

import networkx

from angr.ailment.block import Block
from angr.ailment.expression import (
    Const,
    Insert,
    Phi,
    VirtualVariable,
    VirtualVariableCategory,
    StackBaseOffset,
    Load,
    Convert,
    Expression,
    Tmp,
)
from angr.ailment.manager import Manager
from angr.ailment.statement import Assignment, Store, Return, Jump, ConditionalJump

from angr.knowledge_plugins.functions import Function
from angr.code_location import AILCodeLocation
from angr.analyses import Analysis, register_analysis
from angr.utils.ssa import (
    get_vvar_uselocs,
    get_vvar_deflocs,
    has_ite_expr,
    has_ite_stmt,
    has_tmp_expr,
    is_phi_assignment,
    is_const_assignment,
    is_const_and_vvar_assignment,
    is_const_vvar_load_assignment,
    is_const_vvar_load_dirty_assignment,
    is_const_vvar_tmp_assignment,
    is_vvar_propagatable,
    get_tmp_uselocs,
    get_tmp_deflocs,
    has_store_stmt_in_between_stmts,
)


class SPropagatorModel:
    """
    The SPropagator model that stores replacements for virtual variables.
    """

    def __init__(self):
        self.replacements: Mapping[AILCodeLocation, Mapping[Expression, Expression]] = {}
        # store vvars that are definitely dead (but usually not removed by default because they are stack variables)
        self.dead_vvar_ids: set[int] = set()


class SPropagatorAnalysis(Analysis):
    """
    Constant and expression propagation that only supports SSA AIL graphs.
    """

    def __init__(  # pylint: disable=too-many-positional-arguments
        self,
        subject: Block | Function,
        func_graph: networkx.DiGraph | None = None,
        only_consts: bool = True,
        stack_pointer_tracker=None,
        func_args: set[VirtualVariable] | None = None,
        func_addr: int | None = None,
        ail_manager: Manager | None = None,
        stack_arg_offsets: set[int] | None = None,
    ):
        if isinstance(subject, Block):
            self.block = subject
            self.func = None
            self.mode = "block"
        elif isinstance(subject, Function):
            self.block = None
            self.func = subject
            self.mode = "function"
        else:
            raise TypeError(f"Unsupported subject type {type(subject)}")

        self.func_graph = func_graph
        self.func_addr = func_addr
        self.func_args = func_args
        self.only_consts = only_consts
        self._sp_tracker = stack_pointer_tracker
        self._ail_manager = ail_manager
        self.stack_arg_offsets = stack_arg_offsets

        bp_as_gpr = False
        the_func = None
        if self.func is not None:
            the_func = self.func
        else:
            if self.func_addr is not None:
                with contextlib.suppress(KeyError):
                    the_func = self.kb.functions.get_by_addr(self.func_addr)
        if the_func is not None:
            bp_as_gpr = the_func.info.get("bp_as_gpr", False)
        self._bp_as_gpr = bp_as_gpr

        # output
        self.model = SPropagatorModel()

        self._analyze()

    @property
    def replacements(self):
        return self.model.replacements

    @property
    def dead_vvar_ids(self):
        return self.model.dead_vvar_ids

    def _analyze(self):
        blocks: dict[tuple[int, int | None], Block]
        match self.mode:
            case "block":
                assert self.block is not None
                blocks = {(self.block.addr, self.block.idx): self.block}
            case "function":
                assert self.func_graph is not None
                blocks = {(block.addr, block.idx): block for block in self.func_graph}
            case _:
                raise NotImplementedError

        # find all vvar definitions
        vvar_deflocs = get_vvar_deflocs(blocks.values())
        # find all vvar uses
        vvar_uselocs = get_vvar_uselocs(blocks.values())

        # update vvar_deflocs using function arguments
        if self.func_args:
            for func_arg in self.func_args:
                vvar_deflocs[func_arg.varid] = func_arg, AILCodeLocation.make_extern(func_arg.varid)

        # find all ret sites and indirect jump sites
        retsites: set[tuple[int, int | None, int]] = set()
        jumpsites: set[tuple[int, int | None, int]] = set()
        for bb in blocks.values():
            if bb.statements:
                if isinstance(bb.statements[-1], Return):
                    retsites.add((bb.addr, bb.idx, len(bb.statements) - 1))
                elif isinstance(bb.statements[-1], Jump):
                    jumpsites.add((bb.addr, bb.idx, len(bb.statements) - 1))

        replacements = defaultdict(dict)

        # find constant and other propagatable assignments
        vvarid_to_vvar: dict[int, VirtualVariable] = {}
        const_vvars: dict[int, Const | StackBaseOffset] = {}
        phi_varids: dict[int, set[int | None]] = {}  # mapping from phi_varid to source var IDs
        for vvar_id, (vvar, defloc) in vvar_deflocs.items():
            if defloc.is_extern:
                continue

            assert defloc.block_addr is not None
            assert defloc.stmt_idx is not None

            block = blocks[(defloc.block_addr, defloc.block_idx)]
            stmt = block.statements[defloc.stmt_idx]
            if isinstance(stmt, Assignment) and not (
                isinstance(stmt.dst, VirtualVariable) and stmt.dst.varid == vvar_id
            ):
                # come back later, this is not the def you're looking for
                continue
            if isinstance(stmt, Assignment) and isinstance(stmt.src, Insert):
                # Do not propagate Inserts
                # if this is not acceptable, just make sure we don't proagate inserts into the base of other inserts...
                continue
            if is_phi_assignment(stmt):
                phi_varids[vvar_id] = {
                    src_vvar.varid if src_vvar is not None else None for _, src_vvar in stmt.src.src_and_vvars
                }
            r, v = is_const_assignment(stmt)
            const_value = v
            if r and v is not None and v.tags.get("always_propagate", False):
                # constant propagation
                pass
            elif (
                self.stack_arg_offsets is not None
                and isinstance(stmt, Assignment)
                and isinstance(stmt.dst, VirtualVariable)
                and stmt.dst.was_stack
                and stmt.dst.stack_offset in self.stack_arg_offsets
                and not isinstance(stmt.src, Phi)
            ):
                # force propagation of stack variables to callsites; we set v to stmt.src, but const_value stays None
                r = True
                v = stmt.src
            elif not vvar.was_reg and not vvar.was_parameter:
                continue

            vvarid_to_vvar[vvar_id] = vvar

            if r:
                # replace wherever it's used
                assert v is not None
                if const_value is not None:
                    const_vvars[vvar_id] = const_value
                for vvar_at_use, useloc in vvar_uselocs[vvar_id]:
                    self.replace(replacements, useloc, vvar_at_use, v)

        # function mode only
        if self.mode == "function":
            assert self.func_graph is not None

            # find phi assignments whose source vvars are all constants/stackptrs
            # iterate until it reaches a fixed point
            changed = True
            while changed:
                changed = False
                for vvar_id, src_varids in phi_varids.items():
                    vvar, defloc = vvar_deflocs[vvar_id]
                    if vvar_id in const_vvars:
                        continue
                    if defloc.is_extern:
                        continue

                    assert defloc.block_addr is not None
                    assert defloc.stmt_idx is not None

                    if not vvar.was_reg and not vvar.was_parameter:
                        continue
                    if None in src_varids:
                        continue

                    expanded_src_varids = self._nonphi_src_varids(src_varids, phi_varids)
                    if None not in expanded_src_varids and all(varid in const_vvars for varid in expanded_src_varids):
                        all_int_src_varids: set[int] = {varid for varid in expanded_src_varids if varid is not None}
                        src_values = {
                            (True, v.value) if isinstance((v := const_vvars[varid]), Const) else (False, v.offset)
                            for varid in all_int_src_varids
                        }
                        if len(src_values) == 1:
                            # replace it!
                            const_value = const_vvars[next(iter(all_int_src_varids))]
                            const_vvars[vvar.varid] = const_value
                            for vvar_at_use, useloc in vvar_uselocs[vvar.varid]:
                                self.replace(replacements, useloc, vvar_at_use, const_value)
                            changed = True

            ret_or_jump_sites = retsites | jumpsites
            for vvar_id, (vvar, defloc) in vvar_deflocs.items():
                if vvar_id not in vvar_uselocs:
                    continue
                if vvar_id in const_vvars:
                    continue
                if defloc.is_extern:
                    continue

                vvar_uselocs_set = set(vvar_uselocs[vvar_id])  # deduplicate
                vvar_useloc_to_count: defaultdict[AILCodeLocation, int] = defaultdict(int)
                for _, useloc in vvar_uselocs[vvar_id]:
                    vvar_useloc_to_count[useloc] += 1

                block = blocks[(defloc.block_addr, defloc.block_idx)]
                stmt = block.statements[defloc.stmt_idx]
                if (
                    (vvar.was_reg or vvar.was_parameter)
                    and sum(vvar_useloc_to_count.values()) <= 2
                    and isinstance(stmt, Assignment)
                    and isinstance(stmt.src, Load)
                ):
                    # do we want to propagate this Load expression if it's used for less than twice?
                    # it's often seen in the following pattern, where propagation will be beneficial:
                    #    v0 = Load(...)
                    #    if (!v0) {
                    #       v1 = v0 + 1;
                    #    }
                    can_replace = True
                    for vvar_useloc in vvar_useloc_to_count:
                        if has_store_stmt_in_between_stmts(self.func_graph, blocks, defloc, vvar_useloc):
                            can_replace = False

                    if can_replace:
                        # we can propagate this load because there is no store between its def and use
                        for vvar_used, vvar_useloc in vvar_uselocs_set:
                            self.replace(replacements, vvar_useloc, vvar_used, stmt.src)
                        continue

                if (
                    (vvar.was_reg or vvar.was_stack)
                    and len(vvar_uselocs_set) == 2
                    and isinstance(stmt, Assignment)
                    and not is_phi_assignment(stmt)
                ):
                    # a special case: in a typical switch-case construct, a variable may be used once for comparison
                    # for the default case and then used again for constructing the jump target. we can propagate this
                    # variable for such cases.
                    uselocs = set(vvar_useloc_to_count)
                    if self.is_vvar_used_for_addr_loading_switch_case(uselocs, blocks) and not has_tmp_expr(stmt.src):
                        for vvar_used, vvar_useloc in vvar_uselocs_set:
                            self.replace(replacements, vvar_useloc, vvar_used, stmt.src)
                        # mark the vvar as dead and should be removed
                        self.model.dead_vvar_ids.add(vvar.varid)
                        continue

                if is_vvar_propagatable(vvar, stmt, self.stack_arg_offsets):
                    if len(vvar_uselocs_set) == 1:
                        vvar_used, vvar_useloc = next(iter(vvar_uselocs_set))
                        if (
                            is_const_vvar_load_assignment(stmt)
                            and not has_store_stmt_in_between_stmts(self.func_graph, blocks, defloc, vvar_useloc)
                            and not has_tmp_expr(stmt.src)
                        ):
                            # we can propagate this load because there is no store between its def and use
                            self.replace(replacements, vvar_useloc, vvar_used, stmt.src)
                            continue

                        if is_const_and_vvar_assignment(stmt) and not has_tmp_expr(stmt.src):
                            # if the useloc is a phi assignment statement, ensure that stmt.src is the same as the phi
                            # variable
                            assert vvar_useloc.block_addr is not None
                            assert vvar_useloc.stmt_idx is not None
                            useloc_stmt = blocks[(vvar_useloc.block_addr, vvar_useloc.block_idx)].statements[
                                vvar_useloc.stmt_idx
                            ]
                            if is_phi_assignment(useloc_stmt):
                                if (
                                    isinstance(stmt.src, VirtualVariable)
                                    and stmt.src.oident == useloc_stmt.dst.oident
                                    and stmt.src.category == useloc_stmt.dst.category
                                ):
                                    self.replace(replacements, vvar_useloc, vvar_used, stmt.src)
                            else:
                                self.replace(replacements, vvar_useloc, vvar_used, stmt.src)
                            continue

                    else:
                        if is_const_and_vvar_assignment(stmt):
                            non_exitsite_uselocs = []
                            exitsite_uselocs_to_count = defaultdict(int)
                            for _, loc in vvar_uselocs[vvar_id]:
                                if (loc.block_addr, loc.block_idx, loc.stmt_idx) in ret_or_jump_sites:
                                    exitsite_uselocs_to_count[loc] += 1
                                else:
                                    non_exitsite_uselocs.append(loc)

                            if not exitsite_uselocs_to_count or max(exitsite_uselocs_to_count.values()) == 1:
                                if len(non_exitsite_uselocs) == 1:
                                    # this vvar is used once if we exclude its uses at ret sites or jump sites, and
                                    # used at most once at every exit site. we can propagate it
                                    for vvar_used, vvar_useloc in vvar_uselocs_set:
                                        self.replace(replacements, vvar_useloc, vvar_used, stmt.src)
                                    continue

                                if (
                                    len(set(non_exitsite_uselocs)) == 1
                                    and not has_ite_expr(stmt.src)
                                    and not has_tmp_expr(stmt.src)
                                ):
                                    useloc = non_exitsite_uselocs[0]
                                    assert useloc.block_addr is not None
                                    assert useloc.stmt_idx is not None
                                    useloc_stmt = blocks[(useloc.block_addr, useloc.block_idx)].statements[
                                        useloc.stmt_idx
                                    ]
                                    if stmt.src.depth <= 3 and not has_ite_stmt(useloc_stmt):
                                        # remove duplicate use locs (e.g., if the variable is used multiple times by the
                                        # same statement) - but ensure stmt is simple enough
                                        for vvar_used, vvar_useloc in vvar_uselocs_set:
                                            self.replace(replacements, vvar_useloc, vvar_used, stmt.src)
                                        continue

                # special logic for global variables: if it's used once or multiple times, and the variable is never
                # updated before it's used, we will propagate the load
                if (vvar.was_reg or vvar.was_parameter) and isinstance(stmt, Assignment) and not has_tmp_expr(stmt.src):
                    stmt_src = stmt.src
                    # unpack conversions
                    while isinstance(stmt_src, Convert):
                        stmt_src = stmt_src.operand
                    if (
                        isinstance(stmt_src, Load)
                        and isinstance(stmt_src.addr, Const)
                        and isinstance(stmt_src.addr.value, int)
                    ):
                        gv_updated = False
                        for _vvar_used, vvar_useloc in vvar_uselocs_set:
                            gv_updated |= self.is_global_variable_updated(
                                self.func_graph,
                                blocks,
                                vvar.varid,
                                stmt_src.addr.value,
                                stmt_src.size,
                                defloc,
                                vvar_useloc,
                            )
                        if not gv_updated:
                            for vvar_used, vvar_useloc in vvar_uselocs_set:
                                self.replace(replacements, vvar_useloc, vvar_used, stmt.src)
                            continue

        for vvar_id, uselocs in vvar_uselocs.items():
            vvar = next(iter(uselocs))[0] if vvar_id not in vvarid_to_vvar else vvarid_to_vvar[vvar_id]
            vvar_uselocs_set = set(uselocs)  # deduplicate

            if self._sp_tracker is not None and vvar.category == VirtualVariableCategory.REGISTER:
                if vvar.oident == self.project.arch.sp_offset:
                    sp_bits = (
                        (self.project.arch.registers["sp"][1] * self.project.arch.byte_width)
                        if "sp" in self.project.arch.registers
                        else None
                    )
                    for vvar_at_use, useloc in vvar_uselocs_set:
                        sb_offset = self._sp_tracker.offset_before(useloc.ins_addr, self.project.arch.sp_offset)
                        if sb_offset is not None:
                            idx = None if self._ail_manager is None else self._ail_manager.next_atom()
                            v = StackBaseOffset(idx, self.project.arch.bits, sb_offset)
                            if sp_bits is not None and vvar.bits < sp_bits:
                                # truncation needed
                                v = Convert(None, sp_bits, vvar.bits, False, v)
                            self.replace(replacements, useloc, vvar_at_use, v)
                    continue
                if not self._bp_as_gpr and vvar.oident == self.project.arch.bp_offset:
                    bp_bits = (
                        (self.project.arch.registers["bp"][1] * self.project.arch.byte_width)
                        if "bp" in self.project.arch.registers
                        else None
                    )
                    for vvar_at_use, useloc in vvar_uselocs_set:
                        sb_offset = self._sp_tracker.offset_before(useloc.ins_addr, self.project.arch.bp_offset)
                        if sb_offset is not None:
                            idx = None if self._ail_manager is None else self._ail_manager.next_atom()
                            v = StackBaseOffset(idx, self.project.arch.bits, sb_offset)
                            if bp_bits is not None and vvar.bits < bp_bits:
                                # truncation needed
                                v = Convert(None, bp_bits, vvar.bits, False, v)
                            self.replace(replacements, useloc, vvar_at_use, v)
                    continue

        # find all tmp definitions
        tmp_deflocs = get_tmp_deflocs(blocks.values())
        # find all tmp uses
        tmp_uselocs = get_tmp_uselocs(blocks.values())

        for block_loc, tmp_and_uses in tmp_uselocs.items():
            for tmp_atom, tmp_uses in tmp_and_uses.items():
                # take a look at the definition and propagate the definition if supported
                block = blocks[block_loc]
                tmp_def_stmtidx = tmp_deflocs[block_loc][tmp_atom]

                stmt = block.statements[tmp_def_stmtidx]
                if isinstance(stmt, Assignment):
                    r, v = is_const_assignment(stmt, self.only_consts)
                    if r:
                        # we can propagate it!
                        for tmp_used, tmp_use_stmtidx in tmp_uses:
                            loc = AILCodeLocation(block_loc[0], block_loc[1], tmp_use_stmtidx)
                            self.replace(replacements, loc, tmp_used, stmt.src)
                        continue

                    r = is_const_vvar_tmp_assignment(stmt)
                    if r:
                        # we can propagate it!
                        if isinstance(stmt.src, VirtualVariable):
                            v = const_vvars.get(stmt.src.varid, stmt.src)
                        else:
                            v = stmt.src

                        for tmp_used, tmp_use_stmtidx in tmp_uses:
                            loc = AILCodeLocation(block_loc[0], block_loc[1], tmp_use_stmtidx)
                            self.replace(replacements, loc, tmp_used, v)
                        continue

                    if len(tmp_uses) <= 2 and is_const_vvar_load_dirty_assignment(stmt):
                        for tmp_used, tmp_use_stmtidx in tmp_uses:
                            same_inst = (
                                block.statements[tmp_def_stmtidx].tags["ins_addr"]
                                == block.statements[tmp_use_stmtidx].tags["ins_addr"]
                            )
                            has_store = any(
                                isinstance(stmt_, Store)
                                for stmt_ in block.statements[tmp_def_stmtidx + 1 : tmp_use_stmtidx]
                            )
                            if same_inst or not has_store:
                                # we can propagate this load because either we do not consider memory aliasing problem
                                # within the same instruction (blocks must be originally lifted with
                                # CROSS_INSN_OPT=False), or there is no store between its def and use.
                                loc = AILCodeLocation(block_loc[0], block_loc[1], tmp_use_stmtidx)
                                self.replace(replacements, loc, tmp_used, stmt.src)

        self.model.replacements = replacements

    @staticmethod
    def is_global_variable_updated(
        func_graph, block_dict, varid: int, gv_addr: int, gv_size: int, defloc: AILCodeLocation, useloc: AILCodeLocation
    ) -> bool:
        defblock = block_dict[(defloc.block_addr, defloc.block_idx)]
        useblock = block_dict[(useloc.block_addr, useloc.block_idx)]

        # traverse a graph slice from the def block to the use block and check if the global variable is updated
        seen = {defblock}
        queue = [defblock]
        while queue:
            block = queue.pop(0)

            start_stmt_idx = defloc.stmt_idx if block is defblock else 0  # inclusive
            end_stmt_idx = useloc.stmt_idx if block is useblock else len(block.statements)  # exclusive
            assert start_stmt_idx is not None
            assert end_stmt_idx is not None

            for idx in range(start_stmt_idx, end_stmt_idx):
                stmt = block.statements[idx]
                if isinstance(stmt, Store) and isinstance(stmt.addr, Const):
                    store_addr = stmt.addr.value
                    store_size = stmt.size
                    if gv_addr <= store_addr < gv_addr + gv_size or store_addr <= gv_addr < store_addr + store_size:
                        return True

            if block is useblock:
                continue

            for succ in func_graph.successors(block):
                if succ not in seen:
                    abort_path = False
                    for stmt in succ.statements:
                        if is_phi_assignment(stmt) and any(
                            vvar.varid == varid for _, vvar in stmt.src.src_and_vvars if vvar is not None
                        ):
                            # the virtual variable is no longer live after this point
                            abort_path = True
                            break
                    if abort_path:
                        continue

                    seen.add(succ)
                    queue.append(succ)

        return False

    @staticmethod
    def is_vvar_used_for_addr_loading_switch_case(uselocs: set[AILCodeLocation], blocks) -> bool:
        """
        Check if a virtual variable is used for loading an address in a switch-case construct.

        :param uselocs: The use locations of the virtual variable.
        :param blocks:  All blocks of the current function.
        :return:        True if the virtual variable is used for loading an address in a switch-case construct, False
                        otherwise.
        """

        if len(uselocs) != 2:
            return False

        useloc_0, useloc_1 = list(uselocs)
        block_0 = blocks[(useloc_0.block_addr, useloc_0.block_idx)]
        stmt_0 = block_0.statements[useloc_0.stmt_idx]
        block_1 = blocks[(useloc_1.block_addr, useloc_1.block_idx)]
        stmt_1 = block_1.statements[useloc_1.stmt_idx]

        if isinstance(stmt_0, Jump):
            stmt_0, stmt_1 = stmt_1, stmt_0
            block_0, block_1 = block_1, block_0
        if not isinstance(stmt_0, ConditionalJump) or not isinstance(stmt_1, Jump):
            return False

        # check if stmt_0 jumps to block_1
        if not isinstance(stmt_0.true_target, Const) or not isinstance(stmt_0.false_target, Const):
            return False
        stmt_0_targets = {
            (stmt_0.true_target.value, stmt_0.true_target_idx),
            (stmt_0.false_target.value, stmt_0.false_target_idx),
        }
        return (block_1.addr, block_1.idx) in stmt_0_targets

    @staticmethod
    def replace(
        replacements: dict[AILCodeLocation, dict[VirtualVariable | Tmp, Expression]],
        loc: AILCodeLocation,
        expr: VirtualVariable | Tmp,
        value: Expression,
    ) -> None:
        replacements[loc][expr] = value

    @staticmethod
    def _nonphi_src_varids(
        src_varids: set[int | None], phivar_to_srcvarids: dict[int, set[int | None]]
    ) -> set[int | None]:
        """
        Expand source var IDs to non-phi var IDs.

        :param src_varids:          The source var IDs.
        :param phivar_to_srcvarids: The mapping from phi var IDs to their source var IDs.
        :return:                    The expanded non-phi var IDs.
        """

        # fast path
        if all(varid not in phivar_to_srcvarids for varid in src_varids):
            return src_varids

        result: set[int | None] = set()
        seen: set[int] = set()  # sufficient if we only store phi var IDs here
        queue: list[int | None] = list(src_varids)
        while queue:
            varid = queue.pop()
            if varid is None:
                result.add(None)
                continue
            if varid in seen:
                continue
            if varid not in phivar_to_srcvarids:
                result.add(varid)
                continue
            seen.add(varid)
            queue.extend(vid for vid in phivar_to_srcvarids[varid] if vid is not None and vid not in seen)
        return result


register_analysis(SPropagatorAnalysis, "SPropagator")



from __future__ import annotations
from typing import TYPE_CHECKING

from angr.knowledge_plugins import FunctionManager, Function
from angr.angrdb.models import DbFunction

if TYPE_CHECKING:
    from angr.knowledge_base import KnowledgeBase
    from angr.angrdb.models import DbKnowledgeBase


class FunctionManagerSerializer:
    """
    Serialize/unserialize a function manager and its functions.
    """

    @staticmethod
    def dump(session, db_kb: DbKnowledgeBase, func_manager: FunctionManager):
        """

        :param session:
        :param DbKnowledgeBase db_kb:
        :param FunctionManager func_manager:
        :return:
        """

        # remove all existing functions
        session.query(DbFunction).filter_by(kb=db_kb).delete()

        for func in func_manager.values():
            db_func = DbFunction(
                kb=db_kb,
                addr=func.addr,
                blob=func.serialize(),
            )
            session.add(db_func)

    @staticmethod
    def load(session, db_kb: DbKnowledgeBase, kb: KnowledgeBase):
        """

        :param session:
        :param DbKnowledgeBase db_kb:
        :param KnowledgeBase kb:
        :return:                        A loaded function manager.
        """

        funcs = FunctionManager(kb)

        db_funcs = session.query(DbFunction).filter_by(kb=db_kb)

        for db_func in db_funcs:
            func = Function.parse(db_func.blob, function_manager=funcs, project=kb._project)
            # Mark as dirty so SpillingFunctionDict will save it to LMDB upon eviction.
            func.mark_dirty()
            funcs[func.addr] = func

        funcs.rebuild_callgraph()

        return funcs



from __future__ import annotations
import traceback
import logging
import zlib

from .ansi import Color, BackgroundColor, color, clear

from .testing import is_testing
from angr.utils.formatting import ansi_color_enabled


class Loggers:
    """
    Implements a loggers manager for angr.
    """

    __slots__ = (
        "_loggers",
        "default_level",
        "handler",
        "profiling_enabled",
    )

    def __init__(self, default_level=logging.WARNING):
        self.default_level = default_level
        self._loggers = {}
        self.load_all_loggers()
        self.profiling_enabled = False

        self.handler = logging.StreamHandler()
        self.handler.setFormatter(CuteFormatter(ansi_color_enabled))

        if not is_testing and len(logging.root.handlers) == 0:
            self.enable_root_logger()
            logging.root.setLevel(self.default_level)

    IN_SCOPE = ("angr", "claripy", "cle", "pyvex", "archinfo", "tracer", "driller", "rex", "patcherex", "identifier")

    def load_all_loggers(self):
        """
        A dumb and simple way to conveniently aggregate all loggers.

        Adds attributes to this instance of each registered logger, replacing '.' with '_'
        """
        for name, logger in logging.Logger.manager.loggerDict.items():
            if any(name.startswith(x + ".") or name == x for x in self.IN_SCOPE):
                self._loggers[name] = logger

    def __getattr__(self, k):
        real_k = k.replace("_", ".")
        if real_k in self._loggers:
            return self._loggers[real_k]
        raise AttributeError(k)

    def __dir__(self):
        return list(super().__dir__()) + list(self._loggers.keys())

    def enable_root_logger(self):
        """
        Enable angr's default logger
        """
        logging.root.addHandler(self.handler)

    def disable_root_logger(self):
        """
        Disable angr's default logger
        """
        logging.root.removeHandler(self.handler)

    @staticmethod
    def setall(level):
        for name in logging.Logger.manager.loggerDict:
            logging.getLogger(name).setLevel(level)


class CuteFormatter(logging.Formatter):
    """
    A log formatter that can print log messages with colors.
    """

    __slots__ = ("_should_color",)

    def __init__(self, should_color: bool):
        super().__init__()
        self._should_color: bool = should_color

    def format(self, record: logging.LogRecord):
        name: str = record.name
        level: str = record.levelname
        message: str = record.getMessage()
        name_len: int = len(name)
        lvl_len: int = len(level)
        if self._should_color:
            # Color level
            if record.levelno >= logging.CRITICAL:
                level = color(Color.red, True) + level + clear
                level = color(BackgroundColor.yellow, False) + level + clear
            elif record.levelno >= logging.ERROR:
                level = color(Color.red, False) + level + clear
            elif record.levelno >= logging.WARNING:
                level = color(Color.yellow, False) + level + clear
            elif record.levelno >= logging.INFO:
                level = color(Color.blue, False) + level + clear
            # Color text
            c: int = zlib.adler32(record.name.encode()) % 7
            if c != 0:  # Do not color black or white, allow 'uncolored'
                col = Color(c + Color.black.value)
                message = color(col, False) + message + clear
                name = color(col, False) + name + clear
        # Finalize log message
        name = name.ljust(14 + len(name) - name_len)
        level = level.ljust(8 + len(level) - lvl_len)
        body: str = f"{level} | {self.formatTime(record, self.datefmt): <23} | {name} | {message}"
        if record.exc_info:
            body += "\n" + "".join(traceback.format_exception(*record.exc_info))[:-1]
        return body


def is_enabled_for(logger, level):
    if level == 1:
        from angr import loggers

        return loggers.profiling_enabled
    return originalIsEnabledFor(logger, level)


originalIsEnabledFor = logging.Logger.isEnabledFor

# Override isEnabledFor() for Logger class
logging.Logger.isEnabledFor = is_enabled_for



from __future__ import annotations

import logging
import threading
from typing import overload, TYPE_CHECKING

import archinfo
from archinfo.arch_soot import ArchSoot, SootAddressDescriptor

from angr.exploration_techniques.base import ExplorationTechnique

from .knowledge_plugins.functions import Function
from .sim_state import SimState
from .calling_conventions import default_cc, SimRegArg, SimStackArg, PointerWrapper, SimCCUnknown
from .callable import Callable
from .errors import AngrError
from .engines import UberEngine, ProcedureEngine
from .sim_type import SimTypeFunction, SimTypeInt
from .codenode import HookNode, SyscallNode
from .block import Block, SootBlock
from .sim_manager import SimulationManager

try:
    from .engines import UberEnginePcode
    from .engines.pcode import register_pcode_arch_default_cc
except ImportError:
    UberEnginePcode = None
    register_pcode_arch_default_cc = None

if TYPE_CHECKING:
    from angr import Project, SimCC
    from angr.engines import SimEngine


l = logging.getLogger(name=__name__)


class AngrObjectFactory:
    """
    This factory provides access to important analysis elements.
    """

    project: Project
    default_engine_factory: type[SimEngine]
    procedure_engine: ProcedureEngine
    _default_cc: type[SimCC] | None

    # We use thread local storage to cache engines on a per-thread basis
    _tls: threading.local

    def __init__(self, project, default_engine: type[SimEngine] | None = None):
        self._tls = threading.local()

        if default_engine is None:
            if isinstance(project.arch, archinfo.ArchPcode) and UberEnginePcode is not None:
                l.warning("Creating project with the experimental 'UberEnginePcode' engine")
                self.default_engine_factory = UberEnginePcode
            else:
                self.default_engine_factory = UberEngine
        else:
            self.default_engine_factory = default_engine

        if isinstance(project.arch, archinfo.ArchPcode):
            assert register_pcode_arch_default_cc is not None
            register_pcode_arch_default_cc(project.arch)

        self.project = project
        self._default_cc = default_cc(
            project.arch.name, platform=project.simos.name if project.simos is not None else None, default=SimCCUnknown
        )
        self.procedure_engine = ProcedureEngine(project)

    def __getstate__(self):
        return self.project, self.default_engine_factory, self.procedure_engine, self._default_cc

    def __setstate__(self, state):
        self.project, self.default_engine_factory, self.procedure_engine, self._default_cc = state
        self._tls = threading.local()

    @property
    def default_engine(self):
        if not hasattr(self._tls, "default_engine"):
            self._tls.default_engine = self.default_engine_factory(self.project)
        return self._tls.default_engine

    def snippet(self, addr, jumpkind=None, **block_opts):
        if self.project.is_hooked(addr) and jumpkind != "Ijk_NoHook":
            hook = self.project._sim_procedures[addr]
            size = hook.kwargs.get("length", 0)
            return HookNode(addr, size, self.project.hooked_by(addr))
        if self.project.simos.is_syscall_addr(addr):
            syscall = self.project.simos.syscall_from_addr(addr)
            assert syscall is not None
            size = syscall.kwargs.get("length", 0)
            return SyscallNode(addr, size, syscall)
        return self.block(addr, **block_opts).codenode  # pylint: disable=no-member

    def successors(self, *args, engine=None, **kwargs):
        """
        Perform execution using an engine. Generally, return a SimSuccessors object classifying the results of the run.

        :param state:           The state to analyze
        :param engine:          The engine to use. If not provided, will use the project default.
        :param addr:            optional, an address to execute at instead of the state's ip
        :param jumpkind:        optional, the jumpkind of the previous exit
        :param inline:          This is an inline execution. Do not bother copying the state.

        Additional keyword arguments will be passed directly into each engine's process method.
        """
        if engine is not None:
            return engine.process(*args, **kwargs)
        assert self.default_engine is not None
        return self.default_engine.process(*args, **kwargs)

    def blank_state(self, **kwargs):
        """
        Returns a mostly-uninitialized state object. All parameters are optional.

        :param addr:            The address the state should start at instead of the entry point.
        :param initial_prefix:  If this is provided, all symbolic registers will hold symbolic values with names
                                prefixed by this string.
        :param fs:              A dictionary of file names with associated preset SimFile objects.
        :param concrete_fs:     bool describing whether the host filesystem should be consulted when opening files.
        :param chroot:          A path to use as a fake root directory, Behaves similarly to a real chroot. Used only
                                when concrete_fs is set to True.
        :param kwargs:          Any additional keyword args will be passed to the SimState constructor.
        :return:                The blank state.
        :rtype:                 SimState
        """
        return self.project.simos.state_blank(**kwargs)

    def entry_state(self, **kwargs) -> SimState:
        """
        Returns a state object representing the program at its entry point. All parameters are optional.

        :param addr:            The address the state should start at instead of the entry point.
        :param initial_prefix:  If this is provided, all symbolic registers will hold symbolic values with names
                                prefixed by this string.
        :param fs:              a dictionary of file names with associated preset SimFile objects.
        :param concrete_fs:     boolean describing whether the host filesystem should be consulted when opening files.
        :param chroot:          a path to use as a fake root directory, behaves similar to a real chroot. used only when
                                concrete_fs is set to True.
        :param argc:            a custom value to use for the program's argc. May be either an int or a bitvector. If
                                not provided, defaults to the length of args.
        :param args:            a list of values to use as the program's argv. May be mixed strings and bitvectors.
        :param env:             a dictionary to use as the environment for the program. Both keys and values may be
                                mixed strings and bitvectors.
        :return:                The entry state.
        :rtype:                 SimState
        """

        return self.project.simos.state_entry(**kwargs)

    def full_init_state(self, **kwargs):
        """
        Very much like :meth:`entry_state()`, except that instead of starting execution at the program entry point,
        execution begins at a special SimProcedure that plays the role of the dynamic loader, calling each of the
        initializer functions that should be called before execution reaches the entry point.

        It can take any of the arguments that can be provided to ``entry_state``, except for ``addr``.
        """
        return self.project.simos.state_full_init(**kwargs)

    def call_state(self, addr, *args, **kwargs):
        """
        Returns a state object initialized to the start of a given function, as if it were called with given parameters.

        :param addr:            The address the state should start at instead of the entry point.
        :param args:            Any additional positional arguments will be used as arguments to the function call.

        The following parameters are optional.

        :param base_state:      Use this SimState as the base for the new state instead of a blank state.
        :param cc:              Optionally provide a SimCC object to use a specific calling convention.
        :param ret_addr:        Use this address as the function's return target.
        :param stack_base:      An optional pointer to use as the top of the stack, circa the function entry point
        :param alloc_base:      An optional pointer to use as the place to put excess argument data
        :param grow_like_stack: When allocating data at alloc_base, whether to allocate at decreasing addresses
        :param toc:             The address of the table of contents for ppc64
        :param initial_prefix:  If this is provided, all symbolic registers will hold symbolic values with names
                                prefixed by this string.
        :param fs:              A dictionary of file names with associated preset SimFile objects.
        :param concrete_fs:     bool describing whether the host filesystem should be consulted when opening files.
        :param chroot:          A path to use as a fake root directory, Behaves similarly to a real chroot. Used only
                                when concrete_fs is set to True.
        :param kwargs:          Any additional keyword args will be passed to the SimState constructor.
        :return:                The state at the beginning of the function.
        :rtype:                 SimState

        The idea here is that you can provide almost any kind of python type in `args` and it'll be translated to a
        binary format to be placed into simulated memory. Lists (representing arrays) must be entirely elements of the
        same type and size, while tuples (representing structs) can be elements of any type and size.
        If you'd like there to be a pointer to a given value, wrap the value in a `SimCC.PointerWrapper`. Any value
        that can't fit in a register will be automatically put in a
        PointerWrapper.

        If stack_base is not provided, the current stack pointer will be used, and it will be updated.
        If alloc_base is not provided, the current stack pointer will be used, and it will be updated.
        You might not like the results if you provide stack_base but not alloc_base.

        grow_like_stack controls the behavior of allocating data at alloc_base. When data from args needs to be wrapped
        in a pointer, the pointer needs to point somewhere, so that data is dumped into memory at alloc_base. If you
        set alloc_base to point to somewhere other than the stack, set grow_like_stack to False so that sequential
        allocations happen at increasing addresses.
        """
        return self.project.simos.state_call(addr, *args, **kwargs)

    def simulation_manager(self, thing: list[SimState] | SimState | None = None, **kwargs) -> SimulationManager:
        """
        Constructs a new simulation manager.

        :param thing:           What to put in the new SimulationManager's active stash (either a SimState or a list of
                                SimStates).
        :param kwargs:          Any additional keyword arguments will be passed to the SimulationManager constructor
        :returns:               The new SimulationManager
        :rtype:                 angr.sim_manager.SimulationManager

        Many different types can be passed to this method:

        * If nothing is passed in, the SimulationManager is seeded with a state initialized for the program
          entry point, i.e. :meth:`entry_state()`.
        * If a :class:`SimState` is passed in, the SimulationManager is seeded with that state.
        * If a list is passed in, the list must contain only SimStates and the whole list will be used to seed the
          SimulationManager.
        """
        if thing is None:
            thing = [self.entry_state()]
        elif isinstance(thing, (list, tuple)):
            if any(not isinstance(val, SimState) for val in thing):
                raise AngrError("Bad type to initialize SimulationManager")
        elif isinstance(thing, SimState):
            thing = [thing]
        else:
            raise AngrError(f"BadType to initialize SimulationManager: {thing!r}")

        return SimulationManager(self.project, active_states=thing, **kwargs)

    def simgr(self, *args, **kwargs):
        """
        Alias for `simulation_manager` to save our poor fingers
        """
        return self.simulation_manager(*args, **kwargs)

    def callable(
        self,
        addr: int | Function,
        prototype=None,
        concrete_only=False,
        perform_merge=True,
        base_state=None,
        toc=None,
        cc=None,
        add_options=None,
        remove_options=None,
        techniques: list[ExplorationTechnique] | None = None,
        step_limit: int | None = None,
    ):
        """
        A Callable is a representation of a function in the binary that can be interacted with like a native python
        function.

        :param addr:            The address of the function to use. If you pass in the function object, we will take
                                its addr.
        :param prototype:       The prototype of the call to use, as a string or a SimTypeFunction
        :param concrete_only:   Throw an exception if the execution splits into multiple states
        :param perform_merge:   Merge all result states into one at the end (only relevant if concrete_only=False)
        :param base_state:      The state from which to do these runs
        :param toc:             The address of the table of contents for ppc64
        :param cc:              The SimCC to use for a calling convention
        :param step_limit:      The maximum number of blocks that Callable will execute before pruning the path.
        :returns:               A Callable object that can be used as a interface for executing guest code like a
                                python function.
        :rtype:                 angr.callable.Callable
        """
        if isinstance(addr, Function):
            addr = addr.addr

        return Callable(
            self.project,
            addr=addr,
            prototype=prototype,
            concrete_only=concrete_only,
            perform_merge=perform_merge,
            base_state=base_state,
            toc=toc,
            cc=cc,
            add_options=add_options,
            remove_options=remove_options,
            techniques=techniques,
            step_limit=step_limit,
        )

    def cc(self):
        """
        Return a SimCC (calling convention) parameterized for this project.

        Relevant subclasses of SimFunctionArgument are SimRegArg and SimStackArg, and shortcuts to them can be found on
        this `cc` object.

        For stack arguments, offsets are relative to the stack pointer on function entry.
        """

        assert self._default_cc is not None
        return self._default_cc(arch=self.project.arch)

    def function_prototype(self):
        """
        Return a default function prototype parameterized for this project and SimOS.
        """
        return SimTypeFunction((), SimTypeInt()).with_arch(self.project.arch)

    # pylint: disable=unused-argument, no-self-use, function-redefined
    @overload
    def block(
        self,
        addr: int,
        size=None,
        max_size=None,
        byte_string=None,
        thumb=False,
        backup_state=None,
        extra_stop_points=None,
        opt_level=None,
        num_inst=None,
        traceflags=0,
        insn_bytes=None,
        strict_block_end=None,
        collect_data_refs=False,
        cross_insn_opt=True,
        load_from_ro_regions=False,
        const_prop=False,
        initial_regs=None,
        skip_stmts=False,
    ) -> Block: ...

    # pylint: disable=unused-argument, no-self-use, function-redefined
    @overload
    def block(
        self,
        addr: SootAddressDescriptor,
        size=None,
        max_size=None,
        byte_string=None,
        thumb=False,
        backup_state=None,
        extra_stop_points=None,
        opt_level=None,
        num_inst=None,
        traceflags=0,
        insn_bytes=None,
        strict_block_end=None,
        collect_data_refs=False,
        load_from_ro_regions=False,
        const_prop=False,
        cross_insn_opt=True,
        skip_stmts=False,
    ) -> SootBlock: ...

    def block(
        self,
        addr,
        size=None,
        max_size=None,
        byte_string=None,
        thumb=False,
        backup_state=None,
        extra_stop_points=None,
        opt_level=None,
        num_inst=None,
        traceflags=0,
        insn_bytes=None,
        strict_block_end=None,
        collect_data_refs=False,
        cross_insn_opt=True,
        load_from_ro_regions=False,
        const_prop=False,
        initial_regs=None,
        skip_stmts=False,
    ):
        if isinstance(self.project.arch, ArchSoot) and isinstance(addr, SootAddressDescriptor):
            return SootBlock(addr, arch=self.project.arch, project=self.project)

        if insn_bytes is not None:
            byte_string = insn_bytes

        return Block(
            addr,
            project=self.project,
            size=size,
            max_size=max_size,
            byte_string=byte_string,
            extra_stop_points=extra_stop_points,
            thumb=thumb,
            backup_state=backup_state,
            opt_level=opt_level,
            num_inst=num_inst,
            traceflags=traceflags,
            strict_block_end=strict_block_end,
            collect_data_refs=collect_data_refs,
            cross_insn_opt=cross_insn_opt,
            load_from_ro_regions=load_from_ro_regions,
            const_prop=const_prop,
            initial_regs=initial_regs,
            skip_stmts=skip_stmts,
        )

    def fresh_block(self, addr, size, backup_state=None):
        return Block(addr, project=self.project, size=size, backup_state=backup_state)

    cc.SimRegArg = SimRegArg
    cc.SimStackArg = SimStackArg
    callable.PointerWrapper = PointerWrapper
    call_state.PointerWrapper = PointerWrapper



from __future__ import annotations
import time


import angr

# pylint: disable=arguments-differ,unused-argument


class gettimeofday(angr.SimProcedure):
    def run(self, tv, tz):
        if self.state.solver.is_true(tv == 0):
            return -1

        if angr.options.USE_SYSTEM_TIMES in self.state.options:
            flt = time.time()
            result = {"tv_sec": int(flt), "tv_usec": int(flt * 1000000)}
        else:
            result = {
                "tv_sec": self.state.solver.BVS("tv_sec", self.arch.bits, key=("api", "gettimeofday", "tv_sec")),
                "tv_usec": self.state.solver.BVS("tv_usec", self.arch.bits, key=("api", "gettimeofday", "tv_usec")),
            }

        self.state.mem[tv].struct.timeval = result
        return 0


class clock_gettime(angr.SimProcedure):
    def run(self, which_clock, timespec_ptr):
        if not self.state.solver.is_true(which_clock == 0):
            raise angr.errors.SimProcedureError(
                "clock_gettime doesn't know how to deal with a clock other than CLOCK_REALTIME"
            )

        if self.state.solver.is_true(timespec_ptr == 0):
            return -1

        if angr.options.USE_SYSTEM_TIMES in self.state.options:
            flt = time.time()
            result = {"tv_sec": int(flt), "tv_nsec": int(flt * 1000000000)}
        else:
            result = {
                "tv_sec": self.state.solver.BVS("tv_sec", self.arch.bits, key=("api", "clock_gettime", "tv_sec")),
                "tv_nsec": self.state.solver.BVS("tv_nsec", self.arch.bits, key=("api", "clock_gettime", "tv_nsec")),
            }

        self.state.mem[timespec_ptr].struct.timespec = result
        return 0



# pylint:disable=arguments-differ
from __future__ import annotations
from typing import Literal, Any
from collections import defaultdict

from angr.ailment.expression import Call, Const, StackBaseOffset, VirtualVariable, BinaryOp
from angr.ailment.statement import Assignment, Store, Statement, SideEffectStatement
from angr.ailment.utils import is_none_or_likeable
from angr import SIM_LIBRARIES
from .base import PeepholeOptimizationStmtBase


class ConstAssignmentInfo:
    """
    Holds information about a constant assignment.
    """

    __slots__ = ("base", "bits", "count", "kind", "offset", "value")

    def __init__(
        self, kind: Literal["stack", "heap", "global"], base: Any, offset: int, value: int, bits: int, count: int = 1
    ):
        self.kind: Literal["stack", "heap", "global"] = kind
        self.base = base
        self.offset = offset
        self.value = value
        self.bits = bits
        self.count = count

    def __repr__(self):
        base_str = str(self.base) if self.base is not None else ""
        offset_str = f"+{self.offset}" if self.offset >= 0 else f"{self.offset}"
        return (
            f"ConstAssignmentInfo({self.kind}@{base_str}{offset_str}, {self.value}|{self.bits}bits, count={self.count})"
        )


class InlinedMemset(PeepholeOptimizationStmtBase):
    """
    Simplifies inlined memory setting logic into calls to memset
    """

    MIN_ASSIGNMENTS = 2

    __slots__ = ()

    NAME = "Simplifying inlined memset"
    stmt_classes = (Assignment, Store)

    def optimize(self, stmt: Assignment | Store, stmt_idx: int | None = None, block=None, **kwargs):
        d = self._parse_const_assignment(stmt)
        if d is None:
            return None

        assert block is not None and self.project is not None

        stmt_idx = -1 if stmt_idx is None else stmt_idx
        info: list[tuple[int, ConstAssignmentInfo]] = [(stmt_idx, d)]
        # search forward for constant stores
        i = stmt_idx + 1
        while i < len(block.statements):
            d = self._parse_const_assignment(block.statements[i])
            if d is None:
                break
            info.append((i, d))
            i += 1

        # analyze the collected info
        info_by_kind = defaultdict(list)
        for idx, d in info:
            info_by_kind[d.kind].append((idx, d))

        result = None
        optimized = False
        for _kind, lst in info_by_kind.items():
            if len(lst) <= 1:
                continue
            candidates = self._find_memset_candidates(lst)
            for start_stmt_idx, end_stmt_idx, candidate in candidates:
                if end_stmt_idx - start_stmt_idx <= self.MIN_ASSIGNMENTS:
                    continue

                base_expr = None
                match candidate.kind:
                    case "stack":
                        # base_expr = StackBaseOffset(
                        #     self.manager.next_atom(), self.project.arch.bits, candidate.offset)
                        # Creating memsets on the stack is a recipe to screw up variable identification.
                        # Measurements indicate that memsets on the stack are usually no more than a single
                        # mov or rep stosq. If you have data to the contrary, please help us figure out what
                        # characteristics this has that differentiate it from simply storing many nulls
                        # to many adjacent variables.
                        base_expr = None
                    case "global":
                        base_expr = Const(self.manager.next_atom(), None, candidate.offset, self.project.arch.bits)
                    case "heap":
                        base_expr = BinaryOp(
                            self.manager.next_atom(),
                            "Add",
                            [
                                candidate.base,
                                Const(self.manager.next_atom(), None, candidate.offset, self.project.arch.bits),
                            ],
                            False,
                            bits=self.project.arch.bits,
                        )

                if base_expr is None:
                    continue

                # create a new memset call
                for i in range(start_stmt_idx, end_stmt_idx):
                    block.statements[i] = None  # remove these statements

                assert self.project is not None
                call_stmt = SideEffectStatement(
                    stmt.idx,
                    Call(
                        stmt.idx,
                        "memset",
                        args=[
                            base_expr,
                            Const(self.manager.next_atom(), None, candidate.value, 8),
                            Const(self.manager.next_atom(), None, candidate.count, self.project.arch.bits),
                        ],
                        prototype=SIM_LIBRARIES["libc.so"][0].get_prototype("memset", arch=self.project.arch),
                        **stmt.tags,
                    ),
                    **stmt.tags,
                )
                if start_stmt_idx == stmt_idx:
                    result = call_stmt
                block.statements[start_stmt_idx] = call_stmt
                optimized = True

        if optimized:
            # some optimization has been made
            block.statements = [ss for ss in block.statements if ss is not None]

        return result

    @staticmethod
    def _parse_const_assignment(
        stmt: Statement,
    ) -> ConstAssignmentInfo | None:
        """
        Parse a constant assignment statement to see if it matches the pattern of a memset.

        :param stmt: The statement to parse.
        :return: A tuple of ("stack" | "heap" | "global", base, offset, value, bits) if it matches the pattern, else
                 None.
        """
        if isinstance(stmt, Assignment) and isinstance(stmt.src, Const) and stmt.src.is_int:
            dst = stmt.dst
            if isinstance(dst, VirtualVariable) and dst.was_stack:
                return ConstAssignmentInfo("stack", None, dst.stack_offset, stmt.src.value_int, stmt.src.bits)
        elif isinstance(stmt, Store) and isinstance(stmt.data, Const) and stmt.data.is_int:
            if isinstance(stmt.addr, StackBaseOffset):
                return ConstAssignmentInfo("stack", None, stmt.addr.offset, stmt.data.value_int, stmt.data.bits)
            if isinstance(stmt.addr, Const):
                return ConstAssignmentInfo("global", stmt.addr.value_int, 0, stmt.data.value_int, stmt.data.bits)
            if isinstance(stmt.addr, BinaryOp) and stmt.addr.op in {"Add", "Sub"}:
                base = stmt.addr.operands[0]
                offset_expr = stmt.addr.operands[1]
                if isinstance(offset_expr, Const):
                    offset = offset_expr.value_int if stmt.addr.op == "Add" else -offset_expr.value_int
                    return ConstAssignmentInfo("heap", base, offset, stmt.data.value_int, stmt.data.bits)
            else:
                base = stmt.addr
                return ConstAssignmentInfo("heap", base, 0, stmt.data.value_int, stmt.data.bits)
        return None

    @staticmethod
    def _find_memset_candidates(
        info_lst: list[tuple[int, ConstAssignmentInfo]],
    ) -> list[tuple[int, int, ConstAssignmentInfo]]:
        """
        Analyze a list of ConstAssignmentInfo to find candidates for memset.

        :param info_lst: The list of ConstAssignmentInfo.
        :return: A tuple of (base, start_offset, store_size, value) if they are candidates, else None.
        """
        if len(info_lst) < 2:
            return []

        candidates = []

        info_lst = sorted(info_lst, key=lambda x: x[0])
        start = 0
        while start < len(info_lst) - 1:
            # assignments must be in consecutive statements
            j = start
            while j < len(info_lst):
                if j - start != info_lst[j][0] - info_lst[start][0]:
                    break
                j += 1

            if j - start < 2:
                start += 1
                continue
            end = j

            # assignments from start to j - 1 must cover a continuous memory region
            base = info_lst[start][1].base
            regions = {}
            byte_values = set()
            has_same_base = True

            for j in range(start, end):
                ith_base = info_lst[j][1].base

                if not is_none_or_likeable(ith_base, base):
                    has_same_base = False
                    break
                regions[info_lst[j][1].offset] = info_lst[j][1].bits
                byte_values |= InlinedMemset._int_to_bytes(info_lst[j][1].value, info_lst[j][1].bits)

            if not has_same_base:
                end = j
                if end - start < 2:
                    start += 1
                    continue

            # must have the same byte value
            if len(byte_values) != 1:
                start += 1
                continue

            # last check: are regions continuous?
            next_region_offset = None
            start_region_offset = min(regions)
            has_gap = False
            for region_offset in sorted(regions):
                region_bits = regions[region_offset]
                if next_region_offset is not None and region_offset != next_region_offset:
                    has_gap = True
                    break
                next_region_offset = region_offset + region_bits // 8

            if next_region_offset is None or has_gap:
                start += 1
                continue

            candidate = ConstAssignmentInfo(
                info_lst[start][1].kind,
                base,
                start_region_offset,
                next(iter(byte_values)),
                8,
                next_region_offset - start_region_offset,
            )
            candidates.append(
                (info_lst[start][0], info_lst[end][0] if end < len(info_lst) else info_lst[end - 1][0] + 1, candidate)
            )
            start = end

        return candidates

    @staticmethod
    def _int_to_bytes(value: int, bits: int) -> set[int]:
        """
        Convert an integer value to a set of byte values.

        :param value: The integer value.
        :param bits: The number of bits in the integer.
        :return: A set of byte values.
        """
        byte_values = set()
        num_bytes = bits // 8
        for i in range(num_bytes):
            byte = (value >> (i * 8)) & 0xFF
            byte_values.add(byte)
        return byte_values



from __future__ import annotations
from collections.abc import Generator

from claripy.ast import BV


class AbstractAddressDescriptor:
    """
    AbstractAddressDescriptor describes a list of region+offset tuples. It provides a convenient way for accessing the
    cardinality (the total number of addresses) without enumerating or creating all addresses in static mode.
    """

    __slots__ = ("_regioned_addrs",)

    def __init__(self):
        self._regioned_addrs: list[tuple[str, BV]] = []

    def __len__(self) -> int:
        # this may raise an OverflowError if self.cardinality is greater than sys.maxint
        return self.cardinality

    def __iter__(self) -> Generator[tuple[str, BV]]:
        yield from self._regioned_addrs

    @property
    def cardinality(self):
        n = 0
        for _, si in self._regioned_addrs:
            n += si.cardinality
        return n

    def add_regioned_address(self, region: str, addr: BV):
        self._regioned_addrs.append((region, addr))

    def clear(self):
        self._regioned_addrs.clear()



from __future__ import annotations
import logging

from .paged_memory_mixin import PagedMemoryMixin
from angr.errors import SimSegfaultException, SimMemoryError
from angr.sim_options import STRICT_PAGE_ACCESS

l = logging.getLogger(__name__)


class StackAllocationMixin(PagedMemoryMixin):
    """
    This mixin adds automatic allocation for a stack region based on the stack_end and stack_size parameters.
    """

    # TODO: multiple stacks. this scheme should scale p well
    # TODO tbh this should be handled by an actual fault handler in simos or something
    def __init__(self, stack_end=None, stack_size=None, stack_perms=None, **kwargs):
        super().__init__(**kwargs)
        self._red_pageno = (stack_end - 1) // self.page_size if stack_end is not None else None
        self._remaining_stack = stack_size
        self._stack_perms = stack_perms

    def copy(self, memo):
        o = super().copy(memo)
        o._red_pageno = self._red_pageno
        o._remaining_stack = self._remaining_stack
        o._stack_perms = self._stack_perms
        return o

    def allocate_stack_pages(self, addr: int, size: int, **kwargs):
        """
        Pre-allocates pages for the stack without triggering any logic related to reading from them.

        :param addr: The highest address that should be mapped
        :param size: The number of bytes to be allocated. byte 1 is the one at addr, byte 2 is the one before that, and
                     so on.
        :return: A list of the new page objects
        """
        # weird off-by-ones here. we want to calculate the last byte requested, find its pageno, and then use that to
        # determine what the last page allocated will be and then how many pages are touched
        pageno = addr // self.page_size
        if pageno != self._red_pageno:
            raise SimMemoryError("Trying to allocate stack space in a place that isn't the top of the stack")
        num = pageno - ((addr - size + 1) // self.page_size) + 1

        result = []
        for _ in range(num):
            new_red_pageno = (self._red_pageno - 1) % ((1 << self.state.arch.bits) // self.page_size)
            if new_red_pageno in self._pages:
                raise SimSegfaultException(self._red_pageno * self.page_size, "stack collided with heap")

            if self._remaining_stack is not None and self._remaining_stack < self.page_size:
                raise SimSegfaultException(self._red_pageno * self.page_size, "exhausted stack quota")

            l.debug("Allocating new stack page at %#x", self._red_pageno * self.page_size)
            result.append(
                PagedMemoryMixin._initialize_default_page(
                    self, self._red_pageno, permissions=self._stack_perms, **kwargs
                )
            )
            self._pages[self._red_pageno] = result[-1]

            self._red_pageno = new_red_pageno
            if self._remaining_stack is not None:
                self._remaining_stack -= self.page_size

        return result

    def _initialize_page(self, pageno: int, **kwargs):
        if pageno != self._red_pageno or STRICT_PAGE_ACCESS not in self.state.options:
            return super()._initialize_page(pageno, **kwargs)

        return self.allocate_stack_pages((pageno + 1) * self.page_size - 1, self.page_size)[0]



# pylint:disable=arguments-differ
from __future__ import annotations

from angr.ailment.expression import Const
from angr.ailment.statement import SideEffectStatement

from .base import PeepholeOptimizationStmtBase


class RemoveCxxDestructorCalls(PeepholeOptimizationStmtBase):
    """
    Rewrite C++ operator function calls into operations.
    """

    __slots__ = ()

    NAME = "Remove C++ destructor function calls"
    stmt_classes = (SideEffectStatement,)

    def optimize(self, stmt: SideEffectStatement, **kwargs) -> tuple | None:  # type: ignore
        # are we calling a function that we deem as a C++ destructor?
        assert self.project is not None

        if isinstance(stmt.expr.target, Const):
            func_addr = stmt.expr.target.value
            if not self.project.kb.functions.contains_addr(func_addr):
                return None
            func = self.project.kb.functions[func_addr]
            if "::~" in func.demangled_name and stmt.expr.args is not None:
                # yes it is!
                return ()
        return None



from __future__ import annotations
from collections.abc import Collection
import logging
from typing import Any
from collections import defaultdict

import networkx

from angr.ailment.block import Block

from .structuring.structurer_nodes import ConditionNode, MultiNode

l = logging.getLogger(name=__name__)


class GraphRegion:
    """
    GraphRegion represents a region of nodes.

    :ivar head:             The head of the region.
    :ivar graph:            The region graph.
    :ivar successors:       A set of successors of nodes in the graph. These successors do not belong to the current
                            region.
    :ivar graph_with_successors:    The region graph that includes successor nodes.
    """

    __slots__ = (
        "_node_to_replaced_regions",
        "_replaced_regions",
        "cyclic",
        "cyclic_ancestor",
        "full_graph",
        "graph",
        "graph_with_successors",
        "head",
        "successors",
    )

    def __init__(
        self,
        head,
        graph: networkx.DiGraph[Block | GraphRegion | MultiNode | ConditionNode] | None,
        successors: set[Block | GraphRegion | MultiNode | ConditionNode] | None,
        graph_with_successors: networkx.DiGraph[Block | GraphRegion | MultiNode | ConditionNode] | None,
        cyclic,
        full_graph: networkx.DiGraph[Block] | None,
        cyclic_ancestor: bool = False,
    ):
        self.head = head
        self.graph = graph
        self.successors = set(successors) if successors is not None else None
        # successors inside graph_with_successors should be treated as read-only. when deep-copying GraphRegion objects,
        # successors inside graph_with_successors are *not* deep copied. therefore, you should never modify any
        # successor node in graph_with_successors. to avoid potential programming errors, just treat
        # graph_with_successors as read-only.
        self.graph_with_successors = graph_with_successors

        self.full_graph = full_graph
        self.cyclic = cyclic
        self.cyclic_ancestor = cyclic_ancestor

        # store all subregions and their replacements
        self._node_to_replaced_regions = {}
        self._replaced_regions = {}

    def __repr__(self):
        addrs: list[int] = []
        s = ""
        if self.graph is None:
            # only head is available
            return f"<GraphRegion {self.head!r}>"

        for node in self.graph.nodes():
            if hasattr(node, "addr"):
                assert isinstance(node.addr, int)
                addrs.append(node.addr)
        if addrs:
            s = f": {min(addrs):#x}-{max(addrs):#x}"

        return f"<GraphRegion {self.head!r} of {self.graph.number_of_nodes()} nodes{s}>"

    def copy(self) -> GraphRegion:
        return GraphRegion(
            self.head,
            networkx.DiGraph(self.graph) if self.graph is not None else None,
            set(self.successors) if self.successors is not None else None,
            networkx.DiGraph(self.graph_with_successors) if self.graph_with_successors is not None else None,
            self.cyclic,
            networkx.DiGraph(self.full_graph) if self.full_graph is not None else None,
            cyclic_ancestor=self.cyclic_ancestor,
        )

    def recursive_copy(self, nodes_map=None):
        if nodes_map is None:
            nodes_map = {}
        new_graph = self._recursive_copy(self.graph, nodes_map)

        if self.graph_with_successors is not None:
            assert self.successors is not None
            successors = set()
            for succ in self.successors:
                if succ not in nodes_map:
                    if isinstance(succ, GraphRegion):
                        nodes_map[succ] = succ.recursive_copy(nodes_map=nodes_map)
                    else:
                        nodes_map[succ] = succ
                successors.add(nodes_map[succ])

            new_graph_with_successors = self._recursive_copy(self.graph_with_successors, nodes_map)
        else:
            new_graph_with_successors = None
            successors = None

        new_full_graph = self._recursive_copy(self.full_graph, nodes_map) if self.full_graph is not None else None

        return GraphRegion(
            nodes_map[self.head],
            new_graph,
            successors,
            new_graph_with_successors,
            self.cyclic,
            new_full_graph,
            cyclic_ancestor=self.cyclic_ancestor,
        )

    @staticmethod
    def _recursive_copy(old_graph, nodes_map, ignored_nodes=None) -> networkx.DiGraph:
        new_graph = networkx.DiGraph()

        # make copy of each node and add the mapping from old nodes to new nodes into nodes_map
        for node in old_graph.nodes():
            if node in nodes_map:
                new_graph.add_node(nodes_map[node])
            elif ignored_nodes is not None and node in ignored_nodes:
                # do not copy. use the reference instead
                new_graph.add_node(node)
                # drop it into the nodes_map
                nodes_map[node] = node
            else:
                # make recursive copies
                if type(node) is GraphRegion:
                    new_node = node.recursive_copy(nodes_map=nodes_map)
                    nodes_map[node] = new_node
                elif type(node) is MultiNode:
                    new_node = node.copy()
                    nodes_map[node] = new_node
                else:
                    new_node = node
                    nodes_map[node] = new_node
                new_graph.add_node(new_node)

        # add all edges
        for src, dst, edge_data in old_graph.edges(data=True):
            new_graph.add_edge(nodes_map[src], nodes_map[dst], **edge_data)

        return new_graph

    @property
    def addr(self):
        return self.head.addr

    @staticmethod
    def dbg_get_repr(obj, ident=0):
        return obj.dbg_print(ident=ident) if type(obj) is GraphRegion else " " * ident + str(obj)

    def dbg_print(self, ident=0):
        s = self.dbg_get_repr(self.head, ident=ident) + "\n"

        assert self.graph is not None
        successors = list(self.graph.successors(self.head))
        if len(successors) == 2:
            left_kid, right_kid = successors
            s += (
                " " * ident
                + "if (...) {\n"
                + self.dbg_get_repr(left_kid, ident=ident + 2)
                + "\n"
                + " " * ident
                + "}\n"
                + " " * ident
                + "else if (...) {\n"
                + self.dbg_get_repr(right_kid, ident=ident + 2)
                + "\n"
                + " " * ident
                + "}"
            )
            # TODO: other nodes
        elif len(successors) == 1:
            s += self.dbg_get_repr(successors[0], ident=ident)

        return s

    def replace_region(
        self,
        sub_region: GraphRegion,
        updated_sub_region: GraphRegion,
        replace_with,
        virtualized_edges: set[tuple[Any, Any]],
    ):
        assert self.graph is not None
        assert sub_region.graph_with_successors is not None
        assert sub_region.successors is not None
        assert updated_sub_region.graph_with_successors is not None
        assert updated_sub_region.successors is not None
        if sub_region not in self.graph:
            l.error("The sub-region to replace must be in the current region. Note that this method is not recursive.")
            raise Exception

        if sub_region is self.head:
            self.head = replace_with

        # analyze sub_region and removed_edges to understand which edges in the current region to drop
        removed_edges_count_by_outside_preds = defaultdict(int)
        removed_edges_count_by_outside_succs = defaultdict(int)
        for src, dst in virtualized_edges:
            if src not in sub_region.graph and dst in sub_region.graph:
                # incoming edge
                removed_edges_count_by_outside_preds[src] += 1
            elif src in sub_region.graph and dst not in sub_region.graph:
                # outgoing edge
                removed_edges_count_by_outside_succs[dst] += 1

        edges_to_remove = set()
        for outside_pred, removed_edges_count in removed_edges_count_by_outside_preds.items():
            if (
                outside_pred in sub_region.graph_with_successors
                and sub_region.graph_with_successors.out_degree[outside_pred] == removed_edges_count
            ):
                # all incoming edges from this node are gone. this edge must go away
                edges_to_remove.add((outside_pred, sub_region))
        for outside_succ, removed_edges_count in removed_edges_count_by_outside_succs.items():
            if (
                outside_succ in sub_region.graph_with_successors
                and sub_region.graph_with_successors.in_degree[outside_succ] == removed_edges_count
            ):
                # all outgoing edges to this node are gone. this edge must go away
                edges_to_remove.add((sub_region, outside_succ))

        self._replace_node_in_graph(self.graph, sub_region, replace_with, edges_to_remove)
        if self.graph_with_successors is not None:
            if sub_region.successors != updated_sub_region.successors:
                # some successors are no longer in use - remove them from the graph
                for succ in sub_region.successors:
                    real_succs = list(self.graph_with_successors.successors(sub_region))
                    if succ not in updated_sub_region.successors:
                        # find the corresponding node in graph_with_successors
                        real_succ = next(iter(nn for nn in real_succs if nn.addr == succ.addr), None)
                        if real_succ is not None and real_succ not in self.graph:
                            self.graph_with_successors.remove_edge(sub_region, real_succ)
            self._replace_node_in_graph(self.graph_with_successors, sub_region, replace_with, edges_to_remove)

        self._node_to_replaced_regions[replace_with] = sub_region
        self._replaced_regions[sub_region] = replace_with

    def replace_region_with_region(self, sub_region: GraphRegion, replace_with: GraphRegion):
        assert self.graph is not None
        assert replace_with.successors is not None
        assert replace_with.graph_with_successors is not None
        assert replace_with.graph is not None
        if sub_region not in self.graph:
            l.error("The sub-region to replace must be in the current region. Note that this method is not recursive.")
            raise Exception

        if sub_region is self.head:
            self.head = replace_with.head

        # special case: a successor in replace_with.successors is a normal AIL block while the corresponding
        # successor in self.successors is a graph region (with the AIL block as its head). we handle this case here by
        # creating a new graph_with_successors for the replace_with region
        successor_map = {}
        if self.successors and any(succ not in self.successors for succ in replace_with.successors):
            for succ in replace_with.successors:
                if succ not in self.successors:
                    for succ_ in self.successors:
                        if isinstance(succ_, GraphRegion) and succ_.addr == succ.addr:
                            successor_map[succ] = succ_
        if successor_map:
            replace_with_graph_with_successors = networkx.DiGraph()
            for nn in replace_with.graph_with_successors:
                replace_with_graph_with_successors.add_node(successor_map.get(nn, nn))
            for n0, n1 in replace_with.graph_with_successors.edges:
                n0 = successor_map.get(n0, n0)
                n1 = successor_map.get(n1, n1)
                replace_with_graph_with_successors.add_edge(n0, n1)
        else:
            replace_with_graph_with_successors = replace_with.graph_with_successors

        # if complete_successors is True for RegionIdentifier, replace_with.graph_with_successors may include nodes
        # and edges that are *only* reachable from immediate successors. we will want to remove these nodes and edges,
        # otherwise we may end up structuring the same region twice!
        replace_with_graph_with_successors = self._cleanup_graph_with_successors(
            replace_with.graph, replace_with_graph_with_successors
        )

        self._replace_node_in_graph_with_subgraph(
            self.graph,
            self.successors,
            self.full_graph,
            sub_region,
            replace_with_graph_with_successors,
            replace_with.head,
        )
        if self.graph_with_successors is not None:
            self._replace_node_in_graph_with_subgraph(
                self.graph_with_successors,
                None,
                self.full_graph,
                sub_region,
                replace_with_graph_with_successors,
                replace_with.head,
            )

    @staticmethod
    def _cleanup_graph_with_successors(
        graph: networkx.DiGraph, graph_with_successors: networkx.DiGraph
    ) -> networkx.DiGraph:
        expected_nodes = set(graph)
        for n in list(expected_nodes):
            for succ in graph_with_successors.successors(n):
                expected_nodes.add(succ)
        if all(n in expected_nodes for n in graph_with_successors):
            return graph_with_successors
        return graph_with_successors.subgraph(expected_nodes).to_directed()

    @staticmethod
    def _replace_node_in_graph(graph: networkx.DiGraph, node, replace_with, removed_edges: set):
        in_edges = [(src, dst) for src, dst in graph.in_edges(node) if (src, dst) not in removed_edges]
        out_edges = [(src, dst) for src, dst in graph.out_edges(node) if (src, dst) not in removed_edges]

        graph.remove_node(node)
        graph.add_node(replace_with)

        for src, _ in in_edges:
            if src is node:
                graph.add_edge(replace_with, replace_with)
            else:
                graph.add_edge(src, replace_with)

        for _, dst in out_edges:
            if dst is node:
                graph.add_edge(replace_with, replace_with)
            else:
                graph.add_edge(replace_with, dst)

        assert node not in graph

    def _replace_node_in_graph_with_subgraph(
        self,
        graph: networkx.DiGraph[Block | GraphRegion | MultiNode | ConditionNode],
        known_successors: Collection[Block | GraphRegion | MultiNode | ConditionNode] | None,
        reference_full_graph: networkx.DiGraph[Block] | None,
        node,
        sub_graph: networkx.DiGraph,
        sub_graph_head,
    ):
        in_edges = list(graph.in_edges(node))
        out_edges = list(graph.out_edges(node))

        graph.remove_node(node)

        # FIXME: this is a giant hack to work around the problem that the graph region might have been restructured
        # but not updated in *all* other regions whose .graph_with_successors references this graph region (we only
        # update the parent_region graph right now).
        existing_graph_regions: dict[int, GraphRegion] = {r.addr: r for r in graph if isinstance(r, GraphRegion)}
        for r in sub_graph:
            if isinstance(r, GraphRegion) and r not in graph and r.addr in existing_graph_regions:
                self._replaced_regions[r] = existing_graph_regions[r.addr]

        sub_graph_nodes = [self._replaced_regions.get(nn, nn) for nn in sub_graph.nodes]
        sub_graph_edges = [
            (self._replaced_regions.get(src, src), self._replaced_regions.get(dst, dst)) for src, dst in sub_graph.edges
        ]

        for src, _ in in_edges:
            if src is node:
                graph.add_edge(sub_graph_head, sub_graph_head)
            else:
                graph.add_edge(src, sub_graph_head)

        for _, dst in out_edges:
            if dst is node:
                # ignore all self-loops
                continue
            if known_successors is not None and dst in known_successors:
                continue

            dst_in_subgraph = self._node_to_replaced_regions.get(dst, dst)

            # find the correct source
            if isinstance(dst_in_subgraph, GraphRegion) and dst_in_subgraph not in sub_graph:
                # GraphRegion.successors may not store GraphRegion objects. Instead, the heads of GraphRegion objects
                # are stored.
                dst_head = dst_in_subgraph
                while isinstance(dst_head, GraphRegion) and dst_head not in sub_graph:
                    dst_head = dst_head.head

                if dst_head not in sub_graph:
                    # unexpected: structuring failed and resulted in a bad sub_graph
                    l.warning(
                        "Node %r for node %r is not found in the sub graph at address %#x. Nodes may go missing.",
                        dst_head,
                        dst_in_subgraph,
                        sub_graph_head.addr,
                    )
                    continue

                for src in sub_graph.predecessors(dst_head):
                    graph.add_edge(src, dst)
                # replace the corresponding nodes in sub_graph_nodes and sub_graph_edges
                for i in range(len(sub_graph_nodes)):  # pylint:disable=consider-using-enumerate
                    if sub_graph_nodes[i] is dst_head:
                        sub_graph_nodes[i] = dst
                for i in range(len(sub_graph_edges)):  # pylint:disable=consider-using-enumerate
                    if sub_graph_edges[i][0] is dst_head:
                        sub_graph_edges[i] = (dst, sub_graph_edges[i][1])
                    if sub_graph_edges[i][1] is dst_head:
                        sub_graph_edges[i] = (sub_graph_edges[i][0], dst)
            else:
                if dst_in_subgraph in sub_graph:
                    for src in sub_graph.predecessors(dst_in_subgraph):
                        graph.add_edge(self._replaced_regions.get(src, src), dst)
                elif reference_full_graph is not None and dst_in_subgraph in reference_full_graph:
                    for src in reference_full_graph.predecessors(dst_in_subgraph):
                        if src in graph:
                            graph.add_edge(self._replaced_regions.get(src, src), dst)
                else:
                    # it may happen that the dst node no longer exists in sub_graph or its successors
                    # this is because we have deemed that the dst node is no longer a valid successor for sub_graph
                    pass

        graph.add_nodes_from(sub_graph_nodes)
        graph.add_edges_from(sub_graph_edges)
        # finally, remove all nodes from the graph in known_successors. they are only supposed to be in
        # graph_with_successors.
        if known_successors is not None:
            for nn in known_successors:
                if nn in graph:
                    graph.remove_node(nn)

        assert node not in graph



from __future__ import annotations
import logging
import random

from enum import auto, IntFlag
from collections.abc import Generator

import angr
from angr.analyses import Analysis, AnalysesHub
from angr.knowledge_plugins import Function
from angr.sim_state import SimState

from angr.utils.tagged_interval_map import TaggedIntervalMap

log = logging.getLogger(__name__)
log.setLevel(logging.INFO)


class TraceActions(IntFlag):
    """
    Describe memory access actions.
    """

    WRITE = auto()
    EXECUTE = auto()


class TraceClassifier:
    """
    Classify traces.
    """

    def __init__(self, state: SimState | None = None):
        self.map = TaggedIntervalMap()
        if state:
            self.instrument(state)

    def act_mem_write(self, state) -> None:
        """
        SimInspect callback for memory writes.
        """
        addr = state.solver.eval(state.inspect.mem_write_address)
        length = state.inspect.mem_write_length
        if length is None:
            length = len(state.inspect.mem_write_expr) // state.arch.byte_width
        if not isinstance(length, int):
            length = state.solver.eval(length)
        self.map.add(addr, length, TraceActions.WRITE)

    def act_instruction(self, state) -> None:
        """
        SimInspect callback for instruction execution.
        """
        addr = state.inspect.instruction
        if addr is None:
            log.warning("Symbolic addr")
            return

        # FIXME: Ensure block size is correct
        self.map.add(addr, state.block().size, TraceActions.EXECUTE)

    def instrument(self, state) -> None:
        """
        Instrument `state` for tracing.
        """
        state.inspect.b("mem_write", when=angr.BP_BEFORE, action=self.act_mem_write)
        state.inspect.b("instruction", when=angr.BP_BEFORE, action=self.act_instruction)

    def get_smc_address_and_lengths(self) -> Generator[tuple[int, int]]:
        """
        Evaluate the trace to find which areas of memory were both written to and executed.
        """
        smc_flags = TraceActions.WRITE | TraceActions.EXECUTE
        for addr, size, flags in self.map.irange():
            if (flags & smc_flags) == smc_flags:
                yield (addr, size)

    def determine_smc(self) -> bool:
        """
        Evaluate the trace to find areas of memory that were both written to and executed.
        """
        return any(self.get_smc_address_and_lengths())

    def pp(self):
        for a, b, c in self.map.irange():
            print(f"{a:8x} {b} {c}")


class SelfModifyingCodeAnalysis(Analysis):
    """
    Determine if some piece of code is self-modifying.

    This determination is made by simply executing. If an address is executed
    that is also written to, the code is determined to be self-modifying. The
    determination is stored in the `result` property. The `regions` property
    contains a list of (addr, length) regions that were both written to and
    executed.
    """

    result: bool
    regions: list[tuple[int, int]]

    def __init__(self, subject: None | int | str | Function, max_bytes: int = 0, state: SimState | None = None):
        """
        :param subject: Subject of analysis
        :param max_bytes: Maximum number of bytes from subject address. 0 for no limit (default).
        :param state: State to begin executing from.
        """
        assert self.project.selfmodifying_code

        if subject is None:
            subject = self.project.entry
        if isinstance(subject, str):
            addr = self.project.kb.labels.lookup(subject)
            if addr is None:
                addr = self.project.kb.functions[subject].addr
        elif isinstance(subject, Function):
            addr = subject.addr
        elif isinstance(subject, int):
            addr = subject
        else:
            raise ValueError("Not a supported subject")

        if state is None:
            init_state = self.project.factory.call_state(addr)
        else:
            init_state = state.copy()
            init_state.regs.pc = addr

        init_state.options -= angr.sim_options.simplification

        self._trace_classifier = TraceClassifier(init_state)
        simgr = self.project.factory.simgr(init_state)

        kwargs = {}
        if max_bytes:
            kwargs["filter_func"] = lambda s: (
                "active" if s.solver.eval(addr <= s.regs.pc) and s.solver.eval(s.regs.pc < addr + max_bytes) else "oob"
            )

        # FIXME: Early out on SMC detect
        # FIXME: Configurable step threshold
        # FIXME: Loop analysis

        for n in range(100):
            self._update_progress(n)
            simgr.run(n=3)
            random.shuffle(simgr.active)
            simgr.split(from_stash="active", to_stash=simgr.DROP, limit=10)

        # Classify any out of bound entrypoints
        for state_ in simgr.stashes["oob"]:
            self._trace_classifier.act_instruction(state_)

        self.regions = list(self._trace_classifier.get_smc_address_and_lengths())
        self.result = len(self.regions) > 0


AnalysesHub.register_default("SMC", SelfModifyingCodeAnalysis)



# pylint:disable=missing-class-docstring,missing-function-docstring
"""
Semantic variable naming for boolean flag variables.

This module detects variables used as boolean flags and names them appropriately
(e.g., found, done, flag, ok).
"""

from __future__ import annotations
from typing import TYPE_CHECKING
import logging
from collections import defaultdict

from angr import ailment
from angr.ailment.expression import BinaryOp, UnaryOp, Const
from angr.ailment.statement import Assignment, ConditionalJump
from angr.sim_variable import SimVariable

from .naming_base import ClinicNamingBase

if TYPE_CHECKING:
    pass

l = logging.getLogger(name=__name__)

# Names for boolean flag variables
BOOLEAN_FLAG_NAMES = ["flag", "choice"]


class BooleanNaming(ClinicNamingBase):
    """
    Detects variables used as boolean flags and renames them.

    Boolean flag patterns detected:
    - Variables assigned only 0 or 1
    - Variables used directly in conditions (if (var) or if (!var))
    - Variables compared to 1
    - Variables that are results of comparison operations
    """

    PRIORITY = 80  # Lower priority than other patterns

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._bool_candidates: dict[SimVariable, dict] = {}  # var -> {score, assignments, conditions}
        self._name_counter = 0

    def analyze(self) -> dict[SimVariable, str]:
        """
        Analyze the graph for boolean flag patterns.

        :return: Dictionary mapping SimVariable to new name
        """
        if not self._graph:
            return {}

        # Find variables with boolean-like usage patterns
        self._find_boolean_assignments()
        self._find_boolean_conditions()
        self._find_comparison_results()

        # Score and filter candidates
        self._score_candidates()

        # Assign names to high-confidence boolean variables
        self._assign_boolean_names()

        return self._var_to_new_name

    def _find_boolean_assignments(self) -> None:
        """Find variables that are only assigned 0 or 1."""
        var_assignments: dict[SimVariable, list[int | None]] = defaultdict(list)

        for node in self._graph:
            if not isinstance(node, ailment.Block):
                continue

            for stmt in node.statements:
                if not isinstance(stmt, Assignment):
                    continue

                dst_var = self._get_linked_variable(stmt.dst)
                if dst_var is None:
                    continue

                # Check if assigned a constant 0 or 1
                if isinstance(stmt.src, Const):
                    val = stmt.src.value
                    if val in (0, 1):
                        var_assignments[dst_var].append(val)
                    else:
                        # Assigned a non-boolean constant
                        var_assignments[dst_var].append(None)

                # Check if assigned a comparison result
                elif isinstance(stmt.src, BinaryOp) and self._is_comparison_op(stmt.src.op):
                    var_assignments[dst_var].append(1)  # Comparison results are boolean

                # Check if assigned a logical operation result
                elif isinstance(stmt.src, BinaryOp) and stmt.src.op in ("LogicalAnd", "LogicalOr"):
                    var_assignments[dst_var].append(1)  # Logical ops are boolean

                # Check for negation of another variable
                elif isinstance(stmt.src, UnaryOp) and stmt.src.op == "Not":
                    var_assignments[dst_var].append(1)  # Negation suggests boolean

                else:
                    # Other assignment - might not be boolean
                    var_assignments[dst_var].append(None)

        # Record candidates that are only assigned boolean-like values
        for var, assignments in var_assignments.items():
            if not assignments:
                continue

            # Count how many assignments are boolean-like
            bool_count = sum(1 for a in assignments if a is not None)
            total_count = len(assignments)

            if bool_count > 0:
                self._init_candidate(var)
                # Higher score if all assignments are boolean
                if bool_count == total_count:
                    self._bool_candidates[var]["score"] += 30
                else:
                    self._bool_candidates[var]["score"] += 10 * (bool_count / total_count)
                self._bool_candidates[var]["bool_assignments"] = bool_count

    def _find_boolean_conditions(self) -> None:
        """Find variables used directly in conditions."""
        for node in self._graph:
            if not isinstance(node, ailment.Block):
                continue

            for stmt in node.statements:
                if not isinstance(stmt, ConditionalJump):
                    continue

                cond = stmt.condition

                # Check for direct variable use: if (var)
                var = self._get_linked_variable(cond)
                if var is not None:
                    self._init_candidate(var)
                    self._bool_candidates[var]["score"] += 25
                    self._bool_candidates[var]["direct_condition"] = True
                    continue

                # Check for negation: if (!var)
                if isinstance(cond, UnaryOp) and cond.op == "Not":
                    var = self._get_linked_variable(cond.operand)
                    if var is not None:
                        self._init_candidate(var)
                        self._bool_candidates[var]["score"] += 25
                        self._bool_candidates[var]["direct_condition"] = True
                        continue

                # Check for comparison to 1: if (var != 0)
                if isinstance(cond, BinaryOp) and cond.op in ("CmpEQ", "CmpNE"):
                    op0, op1 = cond.operands
                    var = None

                    if isinstance(op1, Const) and op1.value == 1:
                        var = self._get_linked_variable(op0)
                    elif isinstance(op0, Const) and op0.value == 1:
                        var = self._get_linked_variable(op1)

                    if var is not None:
                        self._init_candidate(var)
                        self._bool_candidates[var]["score"] += 20
                        self._bool_candidates[var]["compared_to_bool"] = True

    def _find_comparison_results(self) -> None:
        """Find variables that store comparison results."""
        for node in self._graph:
            if not isinstance(node, ailment.Block):
                continue

            for stmt in node.statements:
                if not isinstance(stmt, Assignment):
                    continue

                dst_var = self._get_linked_variable(stmt.dst)
                if dst_var is None:
                    continue

                # Check if the source is a comparison
                if isinstance(stmt.src, BinaryOp) and self._is_comparison_op(stmt.src.op):
                    self._init_candidate(dst_var)
                    self._bool_candidates[dst_var]["score"] += 20
                    self._bool_candidates[dst_var]["comparison_result"] = True

    @staticmethod
    def _is_comparison_op(op: str) -> bool:
        """Check if an operation is a comparison."""
        comparison_ops = {"CmpEQ", "CmpNE", "CmpLT", "CmpLE", "CmpGT", "CmpGE"}
        return op in comparison_ops

    def _init_candidate(self, var: SimVariable) -> None:
        """Initialize a candidate entry if not exists."""
        if var not in self._bool_candidates:
            self._bool_candidates[var] = {
                "score": 0,
                "bool_assignments": 0,
                "direct_condition": False,
                "compared_to_bool": False,
                "comparison_result": False,
            }

    def _score_candidates(self) -> None:
        """Apply additional scoring heuristics."""
        for var, info in self._bool_candidates.items():
            # Bonus for small variables (likely to be flags)
            if hasattr(var, "size") and var.size is not None and var.size <= 4:  # 4 bytes or less
                info["score"] += 5

            # Bonus for having multiple boolean indicators
            indicators = sum(
                [
                    info["direct_condition"],
                    info["compared_to_bool"],
                    info["comparison_result"],
                    info["bool_assignments"] > 0,
                ]
            )
            if indicators >= 2:
                info["score"] += 15

    def _assign_boolean_names(self) -> None:
        """Assign names to high-confidence boolean variables."""
        # Sort by score (highest first)
        sorted_candidates = sorted(self._bool_candidates.items(), key=lambda x: -x[1]["score"])

        # Threshold for considering a variable as boolean
        SCORE_THRESHOLD = 25

        for var, info in sorted_candidates:
            if info["score"] < SCORE_THRESHOLD:
                continue

            # Assign a name
            if self._name_counter < len(BOOLEAN_FLAG_NAMES):
                new_name = BOOLEAN_FLAG_NAMES[self._name_counter]
            else:
                new_name = f"flag{self._name_counter}"

            self._var_to_new_name[var] = new_name
            self._name_counter += 1

            l.debug(
                "Identified boolean flag %s (score=%d, direct_cond=%s, cmp_result=%s)",
                var.name,
                info["score"],
                info["direct_condition"],
                info["comparison_result"],
            )



# pylint:disable=no-member,raise-missing-from
from __future__ import annotations
import logging
import json

from collections import defaultdict

from angr.calling_conventions import SimCC, SimCCUsercall, CC_NAMES
from angr.codenode import BlockNode, HookNode, FuncNode
from angr.utils.enums_conv import func_edge_type_to_pb, func_edge_type_from_pb
from angr.sim_type import SimType, SimTypeFunction
from angr.protos import primitives_pb2, function_pb2
from angr.utils.types import make_type_reference

l = logging.getLogger(name=__name__)


class CallingConventionSerializer:
    """
    Serialize/deserialize SimCC classes.
    """

    @staticmethod
    def to_json(cc: SimCC) -> dict:
        if isinstance(cc, SimCCUsercall):
            return {
                "t": "SimCCUsercall",
                # TODO: Deserialize the rest of the fields
            }
        return {"t": cc.__class__.__name__}

    @staticmethod
    def from_json(data: dict, arch) -> SimCC | None:
        cc_type = data.get("t")
        if cc_type == "SimCCUsercall":
            return SimCCUsercall(arch, [], None)  # TODO: Deserialize the rest of the fields
        if cc_type not in CC_NAMES:
            l.warning("Unknown calling convention type %s", cc_type)
            return None
        return CC_NAMES[cc_type](arch)


class FunctionParser:
    """
    The implementation of the serialization methods for the <Function> class.
    """

    @staticmethod
    def serialize(function):
        """
        :return :
        """
        # delayed import
        from .function import Function  # pylint:disable=import-outside-toplevel

        obj = Function._get_cmsg()
        obj.ea = function.addr
        obj.is_entrypoint = False  # TODO: Set this up accordingly
        obj.name = function.name
        obj.is_default_name = function.is_default_name
        obj.is_plt = function.is_plt
        obj.is_syscall = function.is_syscall
        obj.is_simprocedure = function.is_simprocedure
        if function.returning is not None:
            obj.returning = function.returning
        obj.alignment = function.is_alignment
        obj.binary_name = function.binary_name or ""
        obj.normalized = function.normalized
        obj.calling_convention = (
            json.dumps(CallingConventionSerializer.to_json(function.calling_convention)).encode("utf-8")
            if function.calling_convention is not None
            else b""
        )
        if function.prototype is None:
            obj.prototype = b""
        else:
            # convert all named structs in the prototype to typerefs; they will be dereferenced when used
            prototype_ref = make_type_reference(function.prototype)
            obj.prototype = json.dumps(prototype_ref.to_json()).encode("utf-8")
        obj.prototype_libname = (function.prototype_libname or "").encode()
        obj.is_prototype_guessed = function.is_prototype_guessed
        obj.info = function.info.to_json().encode("utf-8") if function.info else b""
        obj.ran_cca = function.ran_cca
        obj.previous_names.extend(function.previous_names)

        for endpoint_type, endpoint_nodes in function.endpoints_with_type.items():
            for node in endpoint_nodes:
                ep = primitives_pb2.Endpoint()
                ep.ea = node.addr
                ep.size = node.size
                match endpoint_type:
                    case "call":
                        ep.type = primitives_pb2.EndpointType.CALL
                    case "return":
                        ep.type = primitives_pb2.EndpointType.RETURN
                    case "transition":
                        ep.type = primitives_pb2.EndpointType.TRANSITION
                    case _:
                        continue
                obj.endpoints.append(ep)

        # signature matched?
        if not function.from_signature:
            obj.matched_from = function_pb2.Function.UNMATCHED
        else:
            if function.from_signature == "flirt":
                obj.matched_from = function_pb2.Function.FLIRT
            else:
                raise ValueError(
                    f"Cannot convert from_signature {function.from_signature} into a SignatureSource enum."
                )

        # blocks
        blocks_list = []
        for _, b in function.code_nodes.items():
            block = primitives_pb2.Block()
            block.ea = b.addr
            block.size = b.size
            if isinstance(b, BlockNode):
                block.bytes = b.bytestr
            blocks_list.append(block)
        obj.blocks.extend(blocks_list)  # pylint:disable=no-member

        block_addrs_set = function.block_addrs_set
        # graph
        edges = []
        external_func_addrs = set()
        external_blocks = []

        for node in function.transition_graph:
            # this is a Block in another function, or just another Function instance.
            if isinstance(node, (FuncNode, HookNode)):
                external_func_addrs.add(node.addr)
            elif node.addr not in block_addrs_set and isinstance(node, BlockNode):
                block = primitives_pb2.Block()
                block.ea = node.addr
                block.size = node.size
                block.bytes = node.bytestr or b""
                external_blocks.append(block)

        TRANSITION_JK = func_edge_type_to_pb("transition")  # default edge type
        for src, dst, data in function.transition_graph.edges(data=True):
            edge = primitives_pb2.Edge()
            edge.src_ea = src.addr
            edge.dst_ea = dst.addr
            edge.jumpkind = TRANSITION_JK
            edge.confirmed = 2  # default value
            for key, value in data.items():
                if key == "type":
                    edge.jumpkind = func_edge_type_to_pb(value)
                elif key == "ins_addr":
                    if value is not None:
                        edge.ins_addr = value
                elif key == "stmt_idx":
                    if value is not None:
                        edge.stmt_idx = value
                elif key == "outside":
                    edge.is_outside = value
                elif key == "confirmed":
                    edge.confirmed = 0 if value is False else 1
                else:
                    l.warning('Unexpected edge data type "%s" encountered during serialization.', key)
            edges.append(edge)
        obj.graph.edges.extend(edges)  # pylint:disable=no-member
        # referenced functions
        obj.external_functions.extend(external_func_addrs)  # pylint:disable=no-member
        obj.external_blocks.extend(external_blocks)  # pylint:disable=no-member

        return obj

    @staticmethod
    def parse_from_cmsg(cmsg, function_manager=None, project=None, meta_only: bool = False):
        """
        :param cmsg: The data to instantiate the <Function> from.

        :return Function:
        """
        # delayed import
        from .function import Function  # pylint:disable=import-outside-toplevel

        proto = SimType.from_json(json.loads(cmsg.prototype.decode("utf-8"))) if cmsg.prototype else None
        if proto is not None:
            if not isinstance(proto, SimTypeFunction):
                l.warning("Unexpected type of function prototype deserialized: %s", type(proto))
                proto = None
            elif project is None:
                proto = None  # we cannot assign an arch-less prototype to a function
            else:
                proto = proto.with_arch(project.arch)

        cc = (
            CallingConventionSerializer.from_json(json.loads(cmsg.calling_convention.decode("utf-8")), project.arch)
            if cmsg.calling_convention and project is not None
            else None
        )

        returning = None
        if cmsg.HasField("returning"):
            returning = cmsg.returning

        obj = Function(
            function_manager,
            cmsg.ea,
            name=cmsg.name,
            is_plt=cmsg.is_plt,
            syscall=cmsg.is_syscall,
            is_simprocedure=cmsg.is_simprocedure,
            returning=returning,
            alignment=cmsg.alignment,
            binary_name=None if not cmsg.binary_name else cmsg.binary_name,
            calling_convention=cc,
            prototype=proto,
            prototype_libname=cmsg.prototype_libname or None,
            is_prototype_guessed=cmsg.is_prototype_guessed,
        )
        obj._project = project
        obj.normalized = cmsg.normalized
        obj.info = json.loads(cmsg.info.decode("utf-8")) if cmsg.info else {}
        obj.is_default_name = cmsg.is_default_name
        obj.ran_cca = cmsg.ran_cca
        obj.previous_names = cmsg.previous_names

        # signature matched?
        if cmsg.matched_from == function_pb2.Function.UNMATCHED:
            obj.from_signature = None
        elif cmsg.matched_from == function_pb2.Function.FLIRT:
            obj.from_signature = "flirt"
        else:
            raise ValueError(f"Cannot convert SignatureSource enum {cmsg.matched_from} to Function.from_signature.")

        if meta_only:
            startpoint_addr = cmsg.ea
            obj.startpoint = (
                HookNode(startpoint_addr, 0, project.hooked_by(startpoint_addr))
                if project and project.is_hooked(startpoint_addr)
                else BlockNode(startpoint_addr, 1)
            )  # the size is incorrect, but it should probably be fine?

            block_addrs_set = set()
            for b in cmsg.blocks:
                block_addrs_set.add(b.ea)
            obj._local_block_addrs = block_addrs_set

            for endpoint in cmsg.endpoints:
                block = BlockNode(endpoint.ea, endpoint.size)
                match endpoint.type:
                    case primitives_pb2.EndpointType.CALL:
                        obj._callout_sites.add(block)
                        obj._add_endpoint(block, "call")
                    case primitives_pb2.EndpointType.RETURN:
                        obj._add_return_site(block)
                    case primitives_pb2.EndpointType.TRANSITION:
                        obj.add_jumpout_site(block)
                    case _:
                        continue

            obj.meta_only = True  # can't be serialized again when evicted from the cache
            obj._dirty = False
            return obj

        # blocks
        blocks = {}
        external_blocks = {}
        for b in cmsg.blocks:
            block = BlockNode(b.ea, b.size, bytestr=b.bytes)
            blocks[block.addr] = block

        for b in cmsg.external_blocks:
            block = BlockNode(b.ea, b.size, bytestr=b.bytes)
            external_blocks[block.addr] = block

        # addresses of referenced functions that are not inside the current function
        external_func_addrs = set(cmsg.external_functions)

        # edges
        edges = {}
        fake_return_edges = defaultdict(list)
        for edge_cmsg in cmsg.graph.edges:
            if edge_cmsg.src_ea in blocks:
                src = blocks[edge_cmsg.src_ea]
            else:
                src = FunctionParser._get_hook_or_func_node(
                    edge_cmsg.src_ea,
                    external_blocks,
                    external_func_addrs,
                    project,
                )

            edge_type = func_edge_type_from_pb(edge_cmsg.jumpkind)
            assert edge_type is not None

            if edge_type == "call":
                dst = FuncNode(edge_cmsg.dst_ea)
            else:
                if edge_cmsg.dst_ea in blocks:
                    dst = blocks[edge_cmsg.dst_ea]
                else:
                    dst = FunctionParser._get_hook_or_func_node(
                        edge_cmsg.dst_ea,
                        external_blocks,
                        external_func_addrs,
                        project,
                    )

            data = {
                "outside": edge_cmsg.is_outside,
                "ins_addr": edge_cmsg.ins_addr,
                "stmt_idx": edge_cmsg.stmt_idx,
            }
            if edge_cmsg.confirmed == 0:
                data["confirmed"] = False
            elif edge_cmsg.confirmed == 1:
                data["confirmed"] = True
            if edge_type == "fake_return":
                fake_return_edges[edge_cmsg.src_ea].append((src, dst, data))
            else:
                edges[(edge_cmsg.src_ea, edge_cmsg.dst_ea, edge_type)] = (src, dst, data)

        added_nodes = set()
        for k, v in edges.items():
            src_addr, dst_addr, edge_type = k
            src, dst, data = v

            outside = data.get("outside", False)
            ins_addr = data.get("ins_addr", None)
            stmt_idx = data.get("stmt_idx", None)
            added_nodes.add(src)
            added_nodes.add(dst)
            if edge_type in ("transition", "exception"):
                obj._transit_to(
                    src,
                    dst,
                    outside=outside,
                    ins_addr=ins_addr,
                    stmt_idx=stmt_idx,
                    is_exception=edge_type == "exception",
                    update_func_block_count=False,  # we will update the block count at the end of this function
                )
            elif edge_type in ("call", "syscall"):
                # find the corresponding fake_ret edge
                fake_ret_edge = next(
                    iter(edge_ for edge_ in fake_return_edges[src_addr] if edge_[1].addr == src.addr + src.size), None
                )
                if dst is None:
                    l.warning(
                        "The destination function %#x does not exist, and it cannot be created since function "
                        "manager is not provided. Please consider passing in a function manager to rebuild this "
                        "graph.",
                        dst_addr,
                    )
                else:
                    if isinstance(dst, (FuncNode, HookNode)):
                        obj._call_to(
                            src,
                            dst,
                            None if fake_ret_edge is None else fake_ret_edge[1],
                            stmt_idx=stmt_idx,
                            ins_addr=ins_addr,
                            return_to_outside=fake_ret_edge is None,
                            syscall=edge_type == "syscall",
                            update_func_block_count=False,  # we will update the block count at the end of this function
                        )
                    if fake_ret_edge is not None:
                        fakeret_src, fakeret_dst, fakeret_data = fake_ret_edge
                        added_nodes.add(fakeret_dst)
                        obj._fakeret_to(
                            fakeret_src,
                            fakeret_dst,
                            confirmed=fakeret_data.get("confirmed"),
                            to_outside=fakeret_data.get("outside", None),
                            update_func_block_count=False,  # we will update the block count at the end of this function
                        )
            elif edge_type == "return":
                obj._return_from_call(
                    src,
                    dst,
                    to_outside=outside,
                    confirm_fakeret=False,
                )
            elif edge_type == "fake_return":
                pass

        for endpoint in cmsg.endpoints:
            if endpoint.ea not in blocks:
                continue
            block = blocks[endpoint.ea]
            match endpoint.type:
                case primitives_pb2.EndpointType.CALL:
                    obj._callout_sites.add(block)
                    obj._add_endpoint(block, "call")
                case primitives_pb2.EndpointType.RETURN:
                    obj._add_return_site(block)
                case primitives_pb2.EndpointType.TRANSITION:
                    obj.add_jumpout_site(block)
                case _:
                    continue

        # add leftover blocks
        for block in blocks.values():
            if block not in added_nodes:
                obj._register_node(True, block)

        obj.update_func_block_count()

        obj._dirty = False

        return obj

    @staticmethod
    def _get_hook_or_func_node(addr, external_blocks: dict, external_func_addrs: set[int], project):
        # should we get a hook node, a func node, or a block in external_blocks?
        if addr in external_func_addrs:
            if project is not None and project.is_hooked(addr):
                # get a hook node instead
                return HookNode(addr, 0, project.hooked_by(addr))
            # get a function node
            return FuncNode(addr)
        if addr in external_blocks:
            # create a block
            return external_blocks[addr]

        raise ValueError(
            f"Unsupported case: The block addr {addr:#x} is not an external function or block. "
            f"This probably indicates a bug in angrdb generation."
        )

    @staticmethod
    def _get_func(addr):
        return FuncNode(addr)



from __future__ import annotations
from . import SimSyscallLibrary
from angr.procedures import SIM_PROCEDURES as P

lib = SimSyscallLibrary()
lib.set_library_names("cgcabi")
lib.add_all_from_dict(P["cgc"])
lib.add_number_mapping_from_dict(
    "cgcabi",
    {
        1: "_terminate",
        2: "transmit",
        3: "receive",
        4: "fdwait",
        5: "allocate",
        6: "deallocate",
        7: "random",
    },
)

lib_tracer = SimSyscallLibrary()
lib_tracer.set_library_names("cgcabi_tracer")
lib_tracer.add_all_from_dict(P["tracer"])



from __future__ import annotations
from typing import Any, TYPE_CHECKING
from collections.abc import Callable

if TYPE_CHECKING:
    from collections import deque


def binary_insert(lst: list | deque, elem: Any, key: Callable, lo: int = 0, hi: int | None = None) -> None:
    """
    Insert an element into a sorted list, and keep the list sorted.

    The major difference from bisect.bisect_left is that this function supports a key method, so user doesn't have
    to create the key array for each insertion.

    :param lst:     The list. Must be pre-ordered.
    :param element: An element to insert into the list.
    :param key:     A method to get the key for each element in the list.
    :param lo:      Lower bound of the search.
    :param hi:      Upper bound of the search.
    :return:        None
    """

    if lo < 0:
        raise ValueError("lo must be a non-negative number")

    if hi is None:
        hi = len(lst)

    while lo < hi:
        mid = (lo + hi) // 2
        if key(lst[mid]) < key(elem):
            lo = mid + 1
        else:
            hi = mid

    lst.insert(lo, elem)



from __future__ import annotations

import logging

from angr import ailment
from .optimization_pass import OptimizationPass, OptimizationPassStage

_l = logging.getLogger(name=__name__)


class MipsGpSettingSimplifier(OptimizationPass):
    """
    Removes $gp-setting statements at the beginning of MIPS functions.
    """

    ARCHES = ["MIPS32", "MIPS64"]
    PLATFORMS = ["linux"]
    STAGE = OptimizationPassStage.AFTER_GLOBAL_SIMPLIFICATION
    NAME = "Remove MIPS $gp-setting statements"
    DESCRIPTION = __doc__.strip()  # type: ignore

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.analyze()

    def _check(self):
        gp_stmt = self._find_gp_setting_stmt()
        if gp_stmt is None:
            return False, {}
        return True, {"gp_stmt": gp_stmt}

    def _analyze(self, cache=None):
        gp_stmt = None
        if cache is not None:
            gp_stmt = cache.get("gp_stmt", None)
        if gp_stmt is None:
            gp_stmt = self._find_gp_setting_stmt()
        if gp_stmt is None:
            return

        block, stmt_idx = gp_stmt
        block_copy = block.copy()
        block_copy.statements.pop(stmt_idx)
        self._update_block(block, block_copy)

    def _find_gp_setting_stmt(self) -> tuple[ailment.Block, int] | None:
        """
        Find the AIL statement that sets the $gp value in the entry block.

        Two patterns are recognized:

        1. Fully constant-propagated: the source is a Const equal to the function's known GP value.
           ``vvar{stack} = Const(gp_value)``

        2. PIC-style computation: the source is a BinaryOp adding/subtracting a constant offset
           to/from the t9 register (which holds the function address on entry in PIC MIPS code).
           ``vvar{stack} = Add(Const(offset), vvar{t9})``

        :return:    A tuple of (block, statement_idx) or None if not found.
        """
        gp_value = self._func.info.get("gp")
        if gp_value is None:
            return None

        first_block = self._get_block(self._func.addr)
        if first_block is None:
            return None

        t9_offset = self.project.arch.registers["t9"][0]

        for idx, stmt in enumerate(first_block.statements):
            if not (
                isinstance(stmt, ailment.Stmt.Assignment)
                and isinstance(stmt.dst, ailment.Expr.VirtualVariable)
                and stmt.dst.was_stack
            ):
                continue

            src = stmt.src

            # Pattern 1: direct constant assignment
            if isinstance(src, ailment.Expr.Const) and src.value == gp_value:
                return first_block, idx

            # Pattern 2: t9 + offset (PIC-style GP computation)
            if isinstance(src, ailment.Expr.BinaryOp) and src.op == "Add" and len(src.operands) == 2:
                op0, op1 = src.operands
                # Either order: Const + VVar{t9} or VVar{t9} + Const
                const_op = None
                vvar_op = None
                for op in (op0, op1):
                    if isinstance(op, ailment.Expr.Const):
                        const_op = op
                    elif isinstance(op, ailment.Expr.VirtualVariable) and op.was_reg and op.reg_offset == t9_offset:
                        vvar_op = op
                if const_op is not None and vvar_op is not None and self._func.addr + const_op.value == gp_value:
                    # Verify the offset matches: func_addr + offset == gp_value
                    return first_block, idx

        return None



from __future__ import annotations
import angr
from angr.sim_type import SimTypeInt


class getuid(angr.SimProcedure):
    # pylint: disable=arguments-differ
    def run(self):
        self.return_type = SimTypeInt(16, True)
        return 1000


class getuid32(angr.SimProcedure):
    # pylint: disable=arguments-differ
    def run(self):
        self.return_type = SimTypeInt(32, True)
        return 1000



from __future__ import annotations
import logging

from claripy import StringS

from angr.engines.soot.values import SimSootValue_StringRef
from angr.procedures.java import JavaSimProcedure

l = logging.getLogger("angr.procedures.java.scanner.nextLine")


class ScannerNextLine(JavaSimProcedure):
    __provides__ = (("java.util.Scanner", "nextLine()"),)

    def run(self, this):  # pylint: disable=arguments-differ,unused-argument
        str_ref = SimSootValue_StringRef(self.state.memory.get_new_uuid())
        self.state.memory.store(str_ref, StringS("scanner_return"))
        # save reference in global dict, so we can easily access it later
        try:
            self.state.globals["java.util.Scanner"].append(str_ref)
        except KeyError:
            self.state.globals["java.util.Scanner"] = [str_ref]
        return str_ref



from __future__ import annotations
import angr


class err(angr.SimProcedure):  # pylint:disable=redefined-builtin
    # pylint:disable=arguments-differ,missing-class-docstring,redefined-builtin

    NO_RET = True

    def run(self, eval, fmt):
        fd = self.state.posix.get_fd(1)
        fprintf = angr.SIM_PROCEDURES["libc"]["fprintf"]
        self.inline_call(fprintf, fd, fmt)  # FIXME: This will not properly replace format strings
        self.exit(eval)



from __future__ import annotations
import claripy

import angr


class sigaction(angr.SimProcedure):
    def run(self, signum, act, oldact):  # pylint:disable=arguments-differ,unused-argument
        # TODO: actually do something
        return claripy.BVV(0, self.arch.sizeof["long"])


class rt_sigaction(angr.SimProcedure):
    def run(self, signum, act, oldact, sigsetsize):  # pylint:disable=arguments-differ,unused-argument
        # TODO: actually do something
        # ...hack
        if self.state.solver.is_true(signum == 33):
            return self.state.libc.ret_errno("EINVAL")
        return claripy.BVV(0, self.arch.sizeof["long"])



from __future__ import annotations
import angr


class _Unwind_Resume(angr.SimProcedure):  # pylint:disable=redefined-builtin
    # pylint:disable=arguments-differ

    NO_RET = True

    def run(self):
        pass



from __future__ import annotations
import os
import sys
from collections.abc import Sequence, Callable

if sys.platform == "win32":
    import colorama  # pylint:disable=import-error


ansi_color_enabled: bool = False


def setup_terminal():
    """
    Check if we are running in a TTY. If so, make sure the terminal supports ANSI escape sequences. If not, disable
    colorized output. Sets global `ansi_color_enabled` to True if colorized output should be enabled by default.
    """
    isatty = (
        hasattr(sys.stdout, "isatty") and sys.stdout.isatty() and hasattr(sys.stderr, "isatty") and sys.stderr.isatty()
    )
    if sys.platform == "win32" and isatty:
        if not isinstance(sys.stdout, colorama.ansitowin32.StreamWrapper):
            colorama.init()

    global ansi_color_enabled  # pylint:disable=global-statement
    # https://no-color.org/
    no_color = os.environ.get("NO_COLOR", "")
    ansi_color_enabled = isatty and not no_color


def ansi_color(s: str, color: str | None) -> str:
    """
    Colorize string `s` by wrapping in ANSI escape sequence for given `color`.

    This function does not consider whether escape sequences are functional or not; it is up to the caller to determine
    if its appropriate. Check global `ansi_color_enabled` value in this module.
    """
    if color is None:
        return s

    codes = {
        "black": "30m",
        "bright_black": "90m",
        "gray": "90m",  # alias 'bright black'
        "blue": "34m",
        "bright_blue": "94m",
        "cyan": "36m",
        "bright_cyan": "96m",
        "green": "32m",
        "bright_green": "92m",
        "magenta": "35m",
        "bright_magenta": "95m",
        "red": "31m",
        "bright_red": "91m",
        "white": "37m",
        "bright_white": "97m",
        "yellow": "33m",
        "bright_yellow": "93m",
    }
    return "\u001b[" + codes[color] + s + "\u001b[0m"


def add_edge_to_buffer(
    buf: Sequence[str],
    ref: Sequence[str],
    start: int,
    end: int,
    formatter: Callable[[str], str] | None = None,
    dashed: bool = False,
    ascii_only: bool | None = None,
):
    """
    Draw an edge by adding Unicode box and arrow glyphs to beginning of each line in a list of lines.

    :param buf: Output buffer, used to render formatted edges.
    :param ref: Reference buffer, used to calculate edge depth.
    :param start: Start line.
    :param end: End line, where arrow points.
    :param formatter: Optional callback function used to format the edge before writing it to output buffer.
    :param dashed: Render edge line dashed instead of solid.
    :param ascii_only: Render edge using ASCII characters only. If unspecified, guess by stdout encoding.
    :return:
    """
    abs_start = min(start, end)
    abs_end = max(start, end)
    max_depth = max(map(len, ref[abs_start : abs_end + 1]))
    descending = start < end

    if ascii_only is None:
        # Guess whether we should only use ASCII characters based on stdout encoding
        ascii_only = getattr(sys.stdout, "encoding", None) != "utf-8"

    if ascii_only:
        chars = {
            "start_cap": "-",
            "start_corner": "+",
            "end_cap": ">",
            "end_corner": "+",
            "horizontal": "+" if dashed else "-",
            "vertical": "+" if dashed else "|",
            "spin": "@ ",
        }
    else:
        chars = {
            "start_cap": "╴",
            "start_corner": "╭" if descending else "╰",
            "end_cap": "▸",
            "end_corner": "╰" if descending else "╭",
            "horizontal": "╌" if dashed else "─",
            "vertical": "╎" if dashed else "│",
            "spin": "⟳ ",
        }

    def handle_line(i, edge_str):
        if formatter is not None:
            edge_str = formatter(edge_str)
        buf[i] = edge_str + buf[i]

    if start == end:
        handle_line(start, chars["spin"])
    else:
        handle_line(
            start, (chars["start_corner"] + chars["horizontal"] * (max_depth - len(ref[start])) + chars["start_cap"])
        )
        handle_line(end, (chars["end_corner"] + chars["horizontal"] * (max_depth - len(ref[end])) + chars["end_cap"]))
        for i in range(abs_start + 1, abs_end):
            handle_line(i, chars["vertical"] + " " * (1 + max_depth - len(ref[i])))



from __future__ import annotations

from angr.ailment import Expr
from angr.engines.vex.claripy.ccall import ARMCondHS, ARMCondLE, ARMCondLO, ARMCondMI, ARMCondNE, ARMCondPL
from angr.engines.vex.claripy.ccall import ARMG_CC_OP_SBB, ARMG_CC_OP_SUB

from .rewriter_base import CCallRewriterBase


class ARMCCallRewriter(CCallRewriterBase):
    """
    Implements VEX ccall rewriter for ARM.
    """

    __slots__ = ()

    def _rewrite(self, ccall: Expr.VEXCCallExpression) -> Expr.Expression | None:
        if ccall.callee == "armg_calculate_condition":
            cond_n_op = ccall.operands[0]

            if isinstance(cond_n_op, Expr.Const):
                concrete_cond_n_op = cond_n_op.value_int
                cond_v = concrete_cond_n_op >> 4
                op_v = concrete_cond_n_op & 0xF
                inv = cond_v & 1

                dep_1 = ccall.operands[1]
                dep_2 = ccall.operands[2]

                if cond_v in {ARMCondHS, ARMCondLO}:
                    # armg_calculate_flag_c
                    if op_v == ARMG_CC_OP_SBB:
                        # dep_1 >= dep_2 if dep_3 == 0 else dep_1 > dep_2,
                        #   and then negate the result if inv == 1
                        dep_3 = ccall.operands[3]
                        if isinstance(dep_3, Expr.Const) and dep_3.value_int == 0:
                            r = Expr.BinaryOp(
                                ccall.idx, "CmpGE" if inv == 0 else "CmpLT", (dep_1, dep_2), False, **ccall.tags
                            )
                        else:
                            r = Expr.BinaryOp(
                                ccall.idx, "CmpGT" if inv == 0 else "CmpLE", (dep_1, dep_2), False, **ccall.tags
                            )
                        return Expr.Convert(None, r.bits, ccall.bits, False, r, **ccall.tags)

                elif cond_v in {ARMCondMI, ARMCondPL}:
                    # armg_calculate_flag_n
                    if op_v == ARMG_CC_OP_SUB:
                        # dep_1 < dep_2,
                        #   and then negate the result if inv == 1
                        r = Expr.BinaryOp(
                            ccall.idx, "CmpLT" if inv == 0 else "CmpGE", (dep_1, dep_2), False, **ccall.tags
                        )
                        return Expr.Convert(None, r.bits, ccall.bits, False, r, **ccall.tags)
                elif cond_v == ARMCondLE:
                    if op_v == ARMG_CC_OP_SUB:
                        # dep_1 <= dep_2,
                        #   and then negate the result if inv == 1
                        r = Expr.BinaryOp(ccall.idx, "CmpLE", (dep_1, dep_2), False, **ccall.tags)
                        return Expr.Convert(None, r.bits, ccall.bits, False, r, **ccall.tags)

                elif cond_v == ARMCondNE:  # noqa: SIM102
                    if op_v == ARMG_CC_OP_SUB:
                        # dep_1 != dep_2,
                        #   and then negate the result if inv == 1
                        r = Expr.BinaryOp(ccall.idx, "CmpNE", (dep_1, dep_2), False, **ccall.tags)
                        return Expr.Convert(None, r.bits, ccall.bits, False, r, **ccall.tags)

        return None
